"""
    ModelParams.SdmParams.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from .BaselineParams import get_num_baseparams
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Models.Stochastic.DispersiveUtils import NUM_SDMCOL_PARAMS
from molass_legacy.Optimizer.BasicOptimizer import AVOID_VANISHING_RATIO

NUM_COL_PARAMS = 6      # N, K, x0, poresize, N0, tI 

def get_common_parameter_names(nc):
    xr_names = ["$h_%d$" % k for k in range(nc)]
    rg_names = ["$R_{g%d}$" % k for k in range(nc)]
    mapping_names = ["$mp_a$", "$mp_b$"]
    uv_names = ["$uh_%d$" % k for k in range(nc)]
    mr_names = ["$mr_a$", "$mr_b$"]
    # seccol_names = ["$t_0$", "$r_p$", "N_0", "$N$", "$me$", "$T$", "mp"]
    seccol_names = ["$N$", "$K$", "$t_0$", "$poresize$", "$N_0$", "$t_I$"]

    return xr_names, rg_names, mapping_names, uv_names, mr_names, seccol_names

class SdmParams:
    def __init__(self, n_components):
        self.logger = logging.getLogger(__name__)
        self.n_components = n_components
        self.num_baseparams = get_num_baseparams()
        self.integral_baseline = self.num_baseparams == 3
        self.t0_upper_bound = get_setting("t0_upper_bound")
        self.use_K = False      # use_K was used for the deprecated stochastic model
        self.estimator = None

        nc = n_components - 1

        self.pos = []
        self.pos.append(0)      # [0] xr_params
        sep = nc
        self.pos.append(sep)    # [1] xr_baseparams
        sep += self.num_baseparams
        self.pos.append(sep)    # [2] rgs
        sep += nc
        self.pos.append(sep)    # [3] mapping
        sep = sep+2
        self.pos.append(sep)    # [4] uv_params
        sep += nc
        self.pos.append(sep)    # [5] uv_baseparams
        sep += 5 + self.num_baseparams
        self.pos.append(sep)    # [6] mappable_range
        sep += 2
        self.pos.append(sep)    # [7] end

        self.num_params = sep   # note that this does not include sec params

        self.logger.info("pos=%s", str(self.pos))

    def get_model_name(self):
        return 'SDM'

    def get_estimator(self, editor, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.Estimators.SdmEstimator
            reload(molass_legacy.Estimators.SdmEstimator)
        from molass_legacy.Estimators.SdmEstimator import SdmEstimator

        estimator = SdmEstimator(editor)
        self.t0_upper_bound = estimator.get_t0_upper_bound()
        self.estimator = estimator
        return estimator

    def set_estimator(self, estimator):
        """
        this is required for the on-the-fly debug in PeakEditor
        in order to avoid repeated computation of colparam_bounds.
        the cause of this neccessity is suspected to be a result of bad design.
        """
        self.logger.info("setting estimator for debugging purposes")
        self.estimator = estimator

    def get_colparam_bounds(self):
        return self.estimator.get_colparam_bounds()

    def split_params(self, n, params):
        if self.num_params != len(params):
            raise ValueError("len(params)=%d != %d, which is calculated from number of components %d." % (len(params), self.num_params, n))

        ret_params = []
        for p, q in zip(self.pos[:-1], self.pos[1:]):
            ret_params.append(params[p:q])

        return ret_params

    def split_params_simple(self, params):
        r = len(params) - self.num_params
        if r == 0:
            decomp_params = params
            seccol_params = None
        else:
            assert r == NUM_COL_PARAMS
            decomp_params = params[:-NUM_COL_PARAMS]
            seccol_params = params[-NUM_COL_PARAMS:]

        self.separate_params = self.split_params(self.n_components, decomp_params) + [seccol_params]
        return self.separate_params

    def split_as_unified_params(self, params):
        return self.split_params_simple(params) + [(None, None)]

    def get_param_bounds(self, params, real_bounds=None):
        init_xr_params, init_xr_baseparams, init_rgs, init_mapping, init_uv_params, init_uv_baseparams, init_mappable_range = self.split_params_simple(params)[0:7]

        xr_h_max = np.max(init_xr_params[:])
        xr_h_min = xr_h_max*AVOID_VANISHING_RATIO
        m_allow = 100

        xr_bounds = []
        for h in init_xr_params:
            xr_bounds.append((xr_h_min, xr_h_max*2))       # h

        for k, v in enumerate(init_xr_baseparams):
            v_allow = max(0.1, abs(v))*0.2          # note that v may be zero
            if self.integral_baseline and k == 2:
                bounds = (max(0, v - v_allow), v + v_allow)
            else:
                bounds = (v - v_allow, v + v_allow)
            xr_bounds.append(bounds)

        rg_bounds = [(rg*0.5, rg*2) for rg in init_rgs]
        a, b = init_mapping
        mapping_bounds = [(a*0.8, a*1.2), (-m_allow, m_allow)]

        uv_h_max = np.max(init_uv_params)
        uv_h_min = uv_h_max*AVOID_VANISHING_RATIO
        uv_bounds = [(uv_h_min, uv_h_max*2) for h in init_uv_params]
        for k, v in enumerate(init_uv_baseparams):
            v_allow = max(0.1, abs(v))*0.2          # note that v may be zero
            if self.integral_baseline and k == 7:
                bounds = (max(0, v - v_allow), v + v_allow)
            else:
                bounds = (v - v_allow, v + v_allow)
            uv_bounds.append(bounds)

        f, t = init_mappable_range
        dx = (t - f)*0.1
        range_bounds = [(f-dx, f+dx), (t-dx, t+dx)]

        # colparam_bounds
        # N, K, x0, poresize, tI
        if real_bounds is None:
            colparam_bounds = self.get_colparam_bounds()
            self.logger.info("got colparam_bounds=%s from the estimator", str(colparam_bounds))
        else:
            # self.estimator may be None in this case
            colparam_bounds = list(real_bounds[-NUM_COL_PARAMS:])
            self.logger.info("got colparam_bounds=%s from real_bounds", str(colparam_bounds))

        self.bounds_lengths = [len(b) for b in [xr_bounds, rg_bounds, mapping_bounds, uv_bounds, range_bounds, colparam_bounds]]
        return xr_bounds + rg_bounds + mapping_bounds + uv_bounds + range_bounds + colparam_bounds

    def make_bounds_mask(self):
        bounds_mask = np.zeros(self.num_params + NUM_COL_PARAMS, dtype=bool)
        n = self.n_components
        nc = n - 1
        bounds_mask[0:nc] = True            # xr_params
        sep = nc + self.num_baseparams

        if self.integral_baseline:
            bounds_mask[sep-1] = True           # xr baseline fouling
            bounds_mask[self.pos[6]-1] = True   # uv aseline fouling

        bounds_mask[sep:sep+nc] = True      # rgs

        return bounds_mask

    def get_rg_start_index(self):
        return self.pos[2]

    def get_mr_start_index(self):
        return self.pos[6]

    def get_parameter_names(self):
        xr_names, rg_names, mapping_names, uv_names, mr_names, seccol_names = get_common_parameter_names(self.n_components - 1)
        xr_basenames = ["$xb_a$", "$xb_b$"]
        if self.num_baseparams == 3:
            xr_basenames += ["$xb_r$"]
        uv_basenames = ["$L$", "$x_0$", "$k$", "$b$", "$s_1$", "$s_2$", "$diffratio$"]
        if self.num_baseparams == 3:
            uv_basenames += ["$ub_r$"]
        return np.array(xr_names + xr_basenames + rg_names + mapping_names + uv_names + uv_basenames + mr_names + seccol_names)

    def get_trans_indeces(self):
        return -4, -3, -2, -1

    def split_get_unified_sec_params(self, params):
        tI, t0, rp, N0, N, me, T, mp = self.split_params_simple(params)[-1]
        # K = N*T
        # m = me + mp
        return tI, t0, rp, N0, N*T, me+mp, N, T, me, mp, None, None, None

    def compute_comformance(self, xr_params, rg_params, seccol_params, poresize_bounds=None):
        # task:
        return 0

    def get_peak_pos_array_list(self, x_array):
        nc = self.n_components - 1
        gr_start = self.get_rg_start_index()
        me = 1.5
        mp = 1.5

        pos_array_list = []
        for params in x_array:
            rg_params = params[gr_start:gr_start+nc]
            N, K, x0, poresize, N0, tI  = params[-NUM_COL_PARAMS:]
            T = K/N
            rho = rg_params/poresize
            rho[rho > 1] = 1
            model_trs = x0 + N*T*(1 - rho)**(me + mp)
            pos_array_list.append(model_trs)

        return np.array(pos_array_list).T

    def get_params_sheet(self, parent, params, dsets, optimizer, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.ModelParams.SdmParamsSheet
            reload(molass_legacy.ModelParams.SdmParamsSheet)
        from .SdmParamsSheet import SdmParamsSheet
        return SdmParamsSheet(parent, params, dsets, optimizer)
    
    def get_paramslider_info(self, devel=True):
        if devel:
            from importlib import reload
            import molass_legacy.ModelParams.SdmSliderInfo
            reload(molass_legacy.ModelParams.SdmSliderInfo)
        from .SdmSliderInfo import SdmSliderInfo
        nc = self.n_components - 1
        return SdmSliderInfo(nc=nc)


"""
    UvCorrector.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.stats import linregress
from molass_legacy.KekLib.ScipyUtils import get_spline

NUM_KNOTS = 8
FIT_TAIL_RATIO = 0.005

class UvCorrector:
    def __init__(self, curve1, curve2):
        self.curve1 = curve1
        self.curve2 = curve2
        self.x = curve2.x
        self.y = curve2.y

    def correction_curve(self, x, *params):
        assert False

    def fit(self, peak_region, debug=False, fig_file=None):
        assert False

    def get_widest_possible_fit_slice(self, peak_region, debug=False):
        x = self.x
        y = self.y
        proportions = [FIT_TAIL_RATIO, 1 - FIT_TAIL_RATIO]
        _, (f, t) = peak_region.get_peak_ends(proportions)

        inside = np.zeros(len(x), dtype=bool)
        inside[f:f+5] = True
        inside[t-4:t+1] = True

        fit_slice = slice(f, t+1)
        outside = np.ones(len(x), dtype=bool)
        outside[fit_slice] = False
        params = []
        for i in [inside, outside]:
            slope, intercept = linregress(x[i], y[i])[0:2]
            params.append((slope, intercept))
        params = np.array(params)
        slope0, slope1 = params[:,0]
        slope_deviation = abs((slope1/slope0 - 1)*len(x)/(y[-1] - y[0]))
        print("slope_deviation=", slope_deviation)
        if True:
            # ret_f, ret_t = 0, len(x)-1
            ret_f, ret_t = f, t
        else:
            ret_f, ret_t = self.find_widest_possible_ends(x, y, f, t, params)

        ret_slice = slice(ret_f, ret_t+1)
        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("get_widest_possible_fit_slice")
                ax.plot(x, y)
                for slope, intercept in params:
                    ax.plot(x, x*slope + intercept)
                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymin, ymax)
                for i, p in enumerate([f, t, ret_f, ret_t]):
                    color = "gray" if i < 2 else "cyan"
                    ax.plot([p, p], [ymin, ymax], ":", color=color)
                fig.tight_layout()
                plt.show()

        return ret_slice

    def find_widest_possible_ends(self, x, y, f, t, params):
        return f, t

    def get_corrected_y(self, params=None):
        if params is None:
            params = self.popt
        slope, intercept = params[0:2]
        params_ = params.copy()
        params_[0:2] = 0
        dy_ = self.correction_curve(self.x, *params_)
        return self.y - dy_

    def get_corrected_ppn(self, sy, k, spp3_extra, debug=False):
        # may be deprecated
        x = self.x
        spline = get_spline(x, sy, num_knots=NUM_KNOTS)
        d_spline = spline.derivative(1)
        sspp = sorted(spp3_extra, key=lambda j: -abs(d_spline(j)))

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            csy = spline(x)

            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("get_corrected_ppn")
                ax.plot(x, sy)
                ax.plot(x, csy, color="cyan")
                for j in sspp:
                    ax.plot(x[j], csy[j], "o", label=str(j))
                ax.legend()
                axt = ax.twinx()
                axt.plot(x, d_spline(x), color="yellow")
                fig.tight_layout()
                plt.show()

        return sspp[0:k]



# coding: utf-8
"""
    Terbium.py

    slightly adapted from
        https://terbium.io/2017/12/matplotlib-3d/
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm

def normalize(arr):
    arr_min = np.min(arr)
    return (arr-arr_min)/(np.max(arr)-arr_min)

def explode(data):
    shape_arr = np.array(data.shape)
    size = shape_arr[:3]*2 - 1
    exploded = np.zeros(np.concatenate([size, shape_arr[3:]]), dtype=data.dtype)
    exploded[::2, ::2, ::2] = data
    return exploded

def expand_coordinates(indices):
    x, y, z = indices
    x[1::2, :, :] += 1
    y[:, 1::2, :] += 1
    z[:, :, 1::2] += 1
    return x, y, z

def plot_cube(cube, angle=320):
    IMG_DIM = 50
    fig = plt.figure(figsize=(30/2.54, 30/2.54))
    ax = fig.add_subplot(111, projection='3d')
    ax.view_init(30, angle)
    ax.set_xlim(right=IMG_DIM*2)
    ax.set_ylim(top=IMG_DIM*2)
    ax.set_zlim(top=IMG_DIM*2)
    draw_voxels(ax, cube)
    plt.show()

def draw_voxels(ax, data):
    data = normalize(data)
    facecolors = cm.viridis(data)
    facecolors[:,:,:,-1] = data
    facecolors = explode(facecolors)
    filled = facecolors[:,:,:,-1] != 0
    x, y, z = expand_coordinates(np.indices(np.array(filled.shape) + 1))
    ax.voxels(x, y, z, filled, facecolors=facecolors, shade=False)



# coding: utf-8
"""
    PreviewResultSaver.py

    Copyright (c) 2018-2021, SAXS Team, KEK-PF
"""
import os
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import set_icon
from molass_legacy.KekLib.TkUtils import rational_geometry
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry, FileEntry
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry
from molass_legacy.KekLib.NumpyUtils import np_savetxt
try:
    import molass_legacy.KekLib.CustomMessageBox as MessageBox
except:
    import OurMessageBox as MessageBox

class PreviewResultSaverDialog(Dialog):
    def __init__(self, parent, dialog):
        self.parent = parent
        self.dialog = dialog
        self.conc_dependence = get_setting('conc_dependence')

    def show(self):
        title = "Preview Data Saver"
        Dialog.__init__( self, self.parent, title, auto_geometry=False, geometry_cb=self.adjust_geometry )

    def adjust_geometry(self):
        rational_geometry(self, self.parent, 0.5, 0.7)

    def body(self, body_frame):
        set_icon( self )

        label_frame = Tk.Frame(body_frame)
        label_frame.pack( padx=50, pady=10 )

        guide = Tk.Label(label_frame, text='Select a folder to save into and press "OK"')
        guide.pack()

        detail_frame = Tk.Frame(body_frame)
        detail_frame.pack( padx=50, pady=10 )

        out_dir = (get_setting( 'analysis_folder' ) + '/preview_results').replace('\\', '/')
        row = 0
        label = Tk.Label( detail_frame, text="Result Folder" )
        label.grid( row=row, column=0, sticky=Tk.W )
        self.folder = Tk.StringVar()
        self.folder.set( out_dir )
        folder_entry = FolderEntry( detail_frame, textvariable=self.folder, width=60 )
        folder_entry.grid( row=row, column=1, sticky=Tk.W, padx=10 )

    def buttonbox( self ):
        box = Tk.Frame(self)
        box.pack()
        w = Tk.Button(box, text="Save", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def validate(self, notify=True, post_fix=""):
        try:
            self.save_data(post_fix=post_fix)
            if notify:
                MessageBox.showinfo("INFO", "Saved successfully.", parent=self)
            ret = 1
        except Exception as exc:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            MessageBox.showerror("ERROR", "Save failed with the following error\n" + etb.last_line(), parent=self)
            ret = 0
        return ret

    def save_data(self, post_fix=""):
        folder = self.folder.get()  + post_fix
        if not os.path.exists(folder):
            mkdirs_with_retry(folder)

        q = self.dialog.q

        self.save_extra_data(folder)

        for k, result in enumerate(self.dialog.solver_results):
            A, B, Z, E, lrf_info, C = result
            Ae, Be, Ze = E
            D = lrf_info.data
            # TODO: errors for A, B
            print('save_data: shapes=',  A.shape, B.shape, C.shape)
            A_data = np.vstack( [q, A, Ae] ).T
            np_savetxt( folder + '/A%d.dat' % k, A_data )
            ignore_bq = self.dialog.get_ignore_bq(k)

            if lrf_info.need_bq():
                B_data = np.vstack( [q, B, Be] ).T
                np_savetxt( folder + '/B%d.dat' % k, B_data )

            np_savetxt( folder + '/M%d.dat' % k, D )
            np_savetxt( folder + '/C%d.dat' % k, C )

    def save_extra_data(self, folder):
        extra_folder = os.path.join(folder, 'extra')
        if not os.path.exists(extra_folder):
            mkdirs_with_retry(extra_folder)

        dialog = self.dialog

        ranges_txt = os.path.join(extra_folder, 'ranges.txt')
        with open(ranges_txt, "w") as fh:
            fh.write(str([prange.get_fromto_list() for prange in dialog.cnv_ranges]) + "\n")

        sd = dialog.sd
        curve = sd.get_xray_curve()
        elution_txt = os.path.join(extra_folder, 'elution.txt')
        np_savetxt(elution_txt, np.array([curve.y]).T)

        angle_txt = os.path.join(extra_folder, 'angle.txt')
        np_savetxt(angle_txt, np.array([sd.qvector]).T)



"""
    MonoporeSecParams.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.BasicUtils import Struct
from molass_legacy._MOLASS.SerialSettings import get_setting
from SecTheory.RetensionTime import (make_initial_guess, compute_retention_time,
                                     estimate_conformance_params,
                                     estimate_conformance_params_fixed_poreexponent,
                                    )
from SecTheory.MonoPore import MEMP_LIMIT, N_LIMIT
from SecTheory.ColumnConstants import SECCONF_LOWER_BOUND, BAD_CONFORMANCE_REDUCE
from molass_legacy._MOLASS.SerialSettings import get_setting

NUM_SEC_PARAMS = 6
MEMP_LOWER = 1e-8
PENALTY_SCALE = 1e8

def initial_guess(xr_params):
    """
    note: this won't improve bounds for rp, m in the succession of "Known Best"
    """
    
    init_sec_params = get_setting("init_sec_params")
    return Struct(params=init_sec_params, bounds=None)

zeros5 = np.zeros(NUM_SEC_PARAMS - 1)

def sec_comformance_impl(t0_upper_bound, xr_params, rg_params, monopore_params, poresize_bounds=None):
    t0, rp, N, me, T, mp = monopore_params
    if poresize_bounds is None:
        poresize_penalty = 0
    else:
        poresize_penalty = min(0, rp - poresize_bounds[0])**2 + max(0, rp - poresize_bounds[1])**2
    conformance = max(0, t0 - t0_upper_bound)**2 + min(0, MEMP_LIMIT - (me+mp))**2 + PENALTY_SCALE*min(0, N_LIMIT - N)**2 + np.sum(np.min([zeros5, monopore_params[1:]], axis=0)**2)
    log_conformance = np.log10(max(MEMP_LOWER, conformance + PENALTY_SCALE*poresize_penalty))
    if log_conformance > 0:
        log_conformance *= BAD_CONFORMANCE_REDUCE
    return max(SECCONF_LOWER_BOUND, log_conformance)

class MonoporeSecParams:
    def __init__(self, poresize, poreexponent):
        self.poresize = poresize
        self.poreexponent = poreexponent

        if poreexponent is None or poreexponent == 0:   # poreexponent == 0 is used since set_setting("poreexponent", None) doesn't seem to work
            self.init_method = initial_guess
            self.conf_method = self.sec_comformance
            self.estm_method = None
            self.nump_adjust = 0
        else:
            assert False

        self.t0_upper_bound = get_setting("t0_upper_bound")

    def sec_comformance(self, xr_params, rg_params, monopore_params, poresize_bounds=None):
        return sec_comformance_impl(self.t0_upper_bound, xr_params, rg_params, monopore_params, poresize_bounds)



# coding: utf-8
"""
    AnimatedHistogram.py

    revised and made into a class from the example at
    https://matplotlib.org/3.1.1/gallery/animation/animated_histogram.html

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
# import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.path as path
import matplotlib.animation as animation
import molass_legacy.KekLib.DebugPlot as plt

class AnimatedHistogram:
    def __init__(self, data, num_bins, num_frames, update_data=False, patch_options=None):
        self.data = data
        self.num_bins = num_bins
        self.update_data = update_data
        if update_data:
            self.increment = None
        else:
            self.increment = len(data)//num_frames
        self.patch_options = patch_options

        hist, bin_edges = np.histogram(data, num_bins)

        left = np.array(bin_edges[:-1])
        right = np.array(bin_edges[1:])
        bottom = np.zeros(len(left))
        top = bottom + hist
        nrects = len(left)

        nverts = nrects * (1 + 3 + 1)
        verts = np.zeros((nverts, 2))
        codes = np.ones(nverts, int) * path.Path.LINETO
        codes[0::5] = path.Path.MOVETO
        codes[4::5] = path.Path.CLOSEPOLY
        verts[0::5, 0] = left
        verts[0::5, 1] = bottom
        verts[1::5, 0] = left
        verts[1::5, 1] = top
        verts[2::5, 0] = right
        verts[2::5, 1] = top
        verts[3::5, 0] = right
        verts[3::5, 1] = bottom

        self.left = left
        self.right = right
        self.bottom = bottom
        self.top = top
        self.verts = verts
        self.codes = codes

    def prepare(self, ax):

        num_patches = 1 if self.update_data else 2
        verts_list = []
        if num_patches == 2:
            verts_list.append(self.verts.copy())    # copy the initial values
        verts_list.append(self.verts)

        for i in range(num_patches):
            """
            create two patches in the non self.update_data mode.
            note that you can't use copy.deepcopy for patches
            because "TransformNode instances can not be copied."
            """
            if self.patch_options:
                kwargs = self.patch_options[i]
            else:
                kwargs = {'facecolor':'green', 'edgecolor':'yellow', 'alpha':0.5}

            barpath = path.Path(verts_list[i], self.codes)
            patch = patches.PathPatch(
                        barpath, **kwargs)
            ax.add_patch(patch)

        self.patch = patch

        ax.set_xlim(self.left[0], self.right[-1])
        ax.set_ylim(self.bottom.min(), self.top.max())

    def reset(self):
        return [self.patch, ]

    def animate(self, i):
        # simulate new data coming in
        if self.update_data:
            data = np.random.randn(len(self.data))
        else:
            data = self.data[0:(i+1)*self.increment]
        hist, bin_edges = np.histogram(data, self.num_bins)
        self.top = self.bottom + hist
        self.verts[1::5, 1] = self.top
        self.verts[2::5, 1] = self.top
        return [self.patch, ]

def demo_():
    data = np.random.randn(1000)
    patch_options = [{'facecolor':None}, {'facecolor':'yellow', 'alpha':0.3}]
    ah = AnimatedHistogram(data, 100, 100, patch_options=patch_options)
    # fig, ax = plt.subplots()  # does not work for DebugPlot
    fig = plt.figure()
    ax = fig.gca()
    ah.prepare(ax)
    ani = animation.FuncAnimation(fig, ah.animate, 100, repeat=True, blit=True, init_func=ah.reset, interval=100)
    fig.tight_layout()
    plt.show()

def demo():
    data = np.random.randn(1000)
    patch_options = [{'facecolor':None}, {'facecolor':'yellow', 'alpha':0.3}]
    # fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14,6))
    fig = plt.figure(figsize=(14,6))
    ax1 = fig.add_subplot(121)
    ax2 = fig.add_subplot(122)
    axes = [ax1, ax2]

    a_hists = []
    for ax in axes:
        ah = AnimatedHistogram(data, 100, 100, patch_options=patch_options)
        ah.prepare(ax)
        a_hists.append(ah)

    hist_patches = [ah.reset()[0] for ah in a_hists]

    def reset():
        return hist_patches

    def animate(i):
        for ah in a_hists:
            ah.animate(i)
        return hist_patches

    ani = animation.FuncAnimation(fig, animate, 100, repeat=True, blit=True, init_func=reset, interval=100)
    fig.tight_layout()
    plt.show()



"""
    SecTheory.SecParamsPlot.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
# import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.ElutionCurveModels import egh
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

def compute_sec_peak_params(rp, tI, t0, P, m, rgs, Npc):
    rhos = rgs/rp
    rhos[rhos > 1] = 1
    trs = t0 + P*(1 - rhos)**m
    sigmas = (trs - tI)/np.sqrt(Npc)
    xlim = trs[0] - 8*sigmas[0], trs[-1] + 8*sigmas[-1]
    x = np.linspace(tI, t0+P, 2000)
    return trs, sigmas, xlim, x

def draw_sec_curve(ax, rp, tI, t0, P, m, rgs, Npc, plot_tI=False):
    ax.set_ylabel("$R_g$")

    trs, sigmas, xlim, x = compute_sec_peak_params(rp, tI, t0, P, m, rgs, Npc)

    r = np.flip(np.arange(0, rp*0.9, 0.1))
    t = t0 + P*np.power(1 - r/rp, m)
    ax.plot(t, r, color="cyan", markersize=1, lw=3, label=r"$t_R = t_0 + P \cdot (1-\rho)^m; \;\; r_p=%.3g, \; m=%.3g$" % (rp, m))

    t_ = t0 + P*np.power(1 - rgs/rp, m)
    ax.plot(t_, rgs, "o", color="red", label="elution points")

    m = np.argmax(sigmas)
    smax = sigmas[m]
    if plot_tI:
        xmin = min(tI - smax*2, xlim[0])
    else:
        xmin = min(t0 - smax*2, xlim[0])
    xmax = max(t0 + P + smax*2, xlim[1])

    ax.set_xlim(xmin, xmax)
    dx = (xmax - xmin)*0.05
    ymin, ymax = ax.get_ylim()
    ymin = min(0, ymin)
    ax.set_ylim(ymin, ymax)
    ty = ymin*0.7 + ymax*0.3
    label_fontsize = 14

    tL = t0 + P
    if plot_tI:
        point_list = [(tI, "$t_I$")]
    else:
        point_list = []

    k = 0
    for px, label in point_list + [(t0, "$t_0$"), (t0 + P, "$t_0 + P$")]:
        ax.plot([px, px], [ymin, ymax], ":")
        dx_ = dx if k < 1 and px < trs[0] - smax*2 else -dx
        ax.annotate(label, xy=(px, 0), xytext=(px + dx_, ty), ha='center',
            arrowprops=dict(arrowstyle="-", color='k'), fontsize=label_fontsize )
        k += 1

    ax.legend(bbox_to_anchor=(0.7, 1), loc="upper center")

def plot_sec_params_impl(fig, ax, optimizer, new_params, lrf_info):
    from Experiment.ColumnTypes import get_columnname

    in_folder = get_in_folder()
    ax.set_title("Simple Exclusion Curve on %s with %s" % (in_folder, get_columnname()), fontsize=20)
    ax.set_ylabel("Intensity")

    if optimizer.rg_curve.__class__.__name__.find("Dummy") >= 0:
        # 'DummyRgCurve' object has no attribute 'segments'
        axt = None
    else:
        axt = ax.twinx()
        axt.grid(False)
    axis_info = [fig, (None, ax, None, axt)]    # fig, (ax1, ax2, ax3, axt)
    optimizer.objective_func(new_params, plot=True, axis_info=axis_info)

    rg_params = optimizer.separate_params[2]
    print("rg_params=", rg_params)

    seccol_params = optimizer.separate_params[7]
    print("seccol_params=", seccol_params)

    Npc, rp, tI, t0, P, m = seccol_params

    if axt is None:
        axt = ax.twinx()
        axt.grid(False)

    draw_sec_curve(axt, rp, tI, t0, P, m, rg_params, Npc)


def plot_sec_params(optimizer, params, lrf_info, devel=True):
    if devel:
        from importlib import reload
        import SecTheory.TriporeExclCurve
        reload(SecTheory.TriporeExclCurve)
        import SecTheory.MonoporeExclCurve
        reload(SecTheory.MonoporeExclCurve)
        import SecTheory.OligoporeExclCurve
        reload(SecTheory.OligoporeExclCurve)
    from SecTheory.TriporeExclCurve import estimate_tripore_exclcurve
    from SecTheory.MonoporeExclCurve import estimate_monopore_exclcurve
    from SecTheory.OligoporeExclCurve import estimate_polypore_exclcurve

    new_params = lrf_info.update_optimizer(optimizer, params)

    with plt.Dp():
        fig = plt.figure(figsize=(20,12))
        gs = GridSpec(3,4)
        ax1 = fig.add_subplot(gs[0,0:3])

        ax2 = fig.add_subplot(gs[1,0:3])
        ax2psd = fig.add_subplot(gs[1,3])

        ax3 = fig.add_subplot(gs[2,0:3])
        ax3psd = fig.add_subplot(gs[2,3])

        plot_sec_params_impl(fig, ax1, optimizer, new_params, lrf_info)
        estimate_monopore_exclcurve(fig, ax2, ax2psd, optimizer, lrf_info, new_params)
        # estimate_tripore_exclcurve(fig, ax3, ax3psd, optimizer, lrf_info, new_params)
        estimate_polypore_exclcurve(fig, ax3, ax3psd, optimizer, lrf_info, new_params)

        xlims = []
        for ax in ax1, ax2, ax3:
            xlims.append(ax.get_xlim())
        xlims = np.array(xlims)
        xmin = np.min(xlims[:,0])
        xmax = np.max(xlims[:,1])
        for ax in ax1, ax2, ax3:
            ax.set_xlim(xmin, xmax)

        fig.tight_layout()
        plt.show()



"""
    Optimizer.BackRunner.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import os
import sys
import numpy as np
import subprocess
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry, is_empty_dir
from molass_legacy.Trimming import save_trimming_txt
from .TheUtils import get_optjob_folder_impl
from .NpSharedMemory import get_shm_singleton
from .SettingsSerializer import serialize_for_optimizer

MAX_NUM_JOBS = 1000

class BackRunner:
    def __init__(self):
        self.optjob_folder = get_optjob_folder_impl()
        self.np_shm = get_shm_singleton()
        self.solver = None

    def get_optjob_folder(self):
        return self.optjob_folder

    def get_work_folder(self):
        optjob_folder = self.optjob_folder
        if not os.path.exists(optjob_folder):
            mkdirs_with_retry(optjob_folder)

        ok = False
        for k in range(MAX_NUM_JOBS):
            work_folder = os.path.join(optjob_folder, '%03d'%k)
            if os.path.exists(work_folder):
                if is_empty_dir(work_folder):
                    ok = True
                    break
            else:
                mkdirs_with_retry(work_folder)
                ok = True
                break

        assert ok

        return work_folder

    def set_work_folder(self, folder):
        # this method is currently used by tester
        self.working_folder = folder

    def run(self, optimizer, init_params, niter=100, seed=1234, work_folder=None, debug=True):
        from .FullOptResult import FILES

        n_components = optimizer.n_components
        class_code = optimizer.__class__.__name__
        composite = optimizer.composite

        # note that, for next trials, which is distiguished from the first trial,
        # init_params is an optimized set of params different from optimizer.init_params

        if work_folder is None:
            folder = self.get_work_folder()
        else:
            # i.e., caller has prepared the folder (may be by get_work_folder())
            folder = work_folder

        self.working_folder = folder
        set_setting("optjob_folder", folder)
        set_setting("optworking_folder", folder)    # unifiy these setting items
        init_params_txt = FILES[2]
        init_params_file = os.path.join(folder, init_params_txt)
        np.savetxt(init_params_file, init_params)

        bounds_txt = FILES[7]
        if optimizer.exports_bounds:
            np.savetxt(os.path.join(folder, bounds_txt), optimizer.real_bounds)

        this_folder = os.path.dirname(os.path.abspath( __file__ ))
        optimizer_py = os.path.join(this_folder, 'optimizer.py')

        in_folder = get_setting('in_folder')

        trimming_txt = FILES[6]
        trimming_file = os.path.join(folder, trimming_txt)
        save_trimming_txt(trimming_file)

        # test_pattern = str(get_setting("test_pattern"))
        test_pattern = "0"      # always set it to "0" to suppress execution-blocking messages

        serialized_str = serialize_for_optimizer()  # "poresize_bounds", "t0_upper_bound"

        np_shm_name = "None" if self.np_shm is None else self.np_shm.name

        from .OptimizerUtils import get_impl_method_name
        nnn = int(self.working_folder[-3:])
        self.solver = get_impl_method_name(nnn)

        self.proc = subprocess.Popen([sys.executable, optimizer_py,
                '-c', class_code,
                '-w', folder,
                '-f', in_folder,
                '-n', str(n_components),
                '-i', init_params_txt,
                '-b', bounds_txt,
                '-d', 'linear',
                '-m', str(niter),
                '-s', str(seed),
                '-r', trimming_txt,
                '-p', serialized_str,
                # '-t', '10',
                '-T', test_pattern,
                '-M', np_shm_name,
                '-S', self.solver,
                ])

    def poll(self):
        return self.proc.poll()

    def getpid(self):
        return self.proc.pid

    def get_callback_txt_path(self):
        return os.path.join(self.working_folder, 'callback.txt')

    def terminate(self):
        self.proc.terminate()

    def revive(self):
        # still working on this method as of 20210616
        nodes = os.listdir(self.optjob_folder)
        for i in range(1, 3):
            last_folder = os.path.join(self.optjob_folder, nodes[-i])
            try:
                self.proc = PopenProxy(nodes[-1])
                break
            except:
                pass

class PopenProxy:
    def __init__(self, folder):
        # still working on this method as of 20210616
        pid_txt = os.path.join(folder, "pid.txt")
        # time check
        # pid
        # active
        callback_txt = os.path.join(folder, "callback.txt")
        # time check



# coding: utf-8
"""
    QuadEghMixtureUtils.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import time
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import matplotlib.animation as animation
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from lmfit import minimize, Parameters
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
import molass_legacy.KekLib.DebugPlot as dplt
from .PairedDataSets import PairedDataSets, WAVELENGTH_POINTS, ANGLE_POINTS
from .ProbDensityUtils import plot_hist_data_list
from .EghMixture import egh_pdf, EghMixture
from .QuadMM import QuadMM
from DataUtils import get_in_folder

def set_consistent_base(ax1, axt):
    ymin1, ymax1 = ax1.get_ylim()
    ymint, ymaxt = axt.get_ylim()
    y_ = ymin1/ymax1*ymaxt
    axt.set_ylim(y_, ymaxt)

def get_components(x, y, gmm):
    gy_list = []
    ty = np.zeros(len(x))
    for k in range(len(gmm.pi)):
        w = gmm.pi[k]
        m = gmm.tR[k]
        s = gmm.sigma[k]
        t = gmm.tau[k]
        print([k], w, m, s, t)
        gy = egh_pdf(x, m, s, t, w)
        gy_list.append([gy, m])
        ty += gy

    sorted_gy_list = sorted(gy_list, key=lambda x:x[1])

    if False:
        dplt.push()
        fig = dplt.figure()
        ax = fig.gca()
        ax.plot(ty)
        for k, rec in enumerate(sorted_gy_list):
            gy = rec[0]
            ax.plot(gy, ':', label=str(k))

        ax.legend()
        fig.tight_layout()
        dplt.show
        dplt.pop()

    def obj_func(params):
        S   = params['S']
        return ty*S - y

    params = Parameters()
    S_init = np.max(y)/np.max(ty)
    params.add('S', value=S_init, min=0, max=S_init*100 )
    result = minimize(obj_func, params, args=())

    scale = result.params['S'].value
    return [ scale*rec[0] for rec in sorted_gy_list ], ty*scale

def plot_components(ax, x, cgys):
    ret_lines = []
    for k, gy in enumerate(cgys):
        line, = ax.plot(x, gy, ':', label='c-%d' % k)
        ret_lines.append(line)
    return ret_lines

def plot_decompositions(y_list, data_list, k, max_iter=100, axes=None, titles=None, plot_class=plt, anim_data=False):
    from DataUtils import get_in_folder

    states = np.random.randint(1000, 9999, 4)
    print('random_states:', states)
    qmm = QuadMM(EghMixture, k, max_iter=max_iter, random_states=states, anim_data=anim_data)
    qmm.unified_fit(X=[np.expand_dims(data,1) for data in data_list], bins=[len(y) for y in y_list])

    if axes is None:
        fig, axes = plot_class.subplots(nrows=1, ncols=4, figsize=(20,4))
        fig.suptitle("Decomposition of %s using Dual EGH Mixture Model" % get_in_folder(), fontsize=20)

    mm_list = []
    xy_list = []
    for n, (ax, y, data) in enumerate(zip(axes, y_list, data_list)):
        if titles is not None:
            ax.set_title(titles[n], fontsize=16)

        x = np.arange(len(y))
        ax.plot(x, y, label='input')
        mm = qmm.get_sub(n)
        mm_list.append(mm)
        xy_list.append((x, y))
        cgys, gy = get_components(x, y, mm)
        ax.plot(gy, label='total')
        plot_components(ax, x, cgys)
        ax.legend()

    if axes is None:
        fig.tight_layout()
        fig.subplots_adjust(top=0.88)

    return mm_list, xy_list

def get_anim_components(x, y, gmm, n):
    gy_list = []
    ty = np.zeros(len(x))
    for k in range(len(gmm.pi)):
        w = gmm.pi_array[n,k]
        m = gmm.tR_array[n,k]
        s = gmm.sigma_array[n,k]
        t = gmm.tau_array[n,k]
        # print([n,k], w, m, s, t)
        gy = egh_pdf(x, m, s, t, w)
        gy_list.append([gy, m])
        ty += gy

    sorted_gy_list = sorted(gy_list, key=lambda x:x[1])

    if True:
        dplt.push()
        fig = dplt.figure()
        ax = fig.gca()
        ax.plot(ty)
        for k, rec in enumerate(sorted_gy_list):
            gy = rec[0]
            ax.plot(gy, ':', label=str(k))

        ax.legend()
        fig.tight_layout()
        dplt.show
        dplt.pop()

    def obj_func(params):
        S   = params['S']
        return ty*S - y

    params = Parameters()
    S_init = np.max(y)/np.max(ty)
    params.add('S', value=S_init, min=0, max=S_init*100 )
    result = minimize(obj_func, params, args=())

    scale = result.params['S'].value
    return [ scale*rec[0] for rec in sorted_gy_list ], ty*scale

def animate_gmm_elution(y1, y2, gmm1, gmm2, plot_class=plt):
    from DataUtils import get_in_folder

    # fig, axes = plot_class.subplots(nrows=1, ncols=2, figsize=(14,7))     # does not work for animation
    # ax1, ax2 = axes

    fig = plot_class.figure(figsize=(14,7))
    ax1 = fig.add_subplot(121)
    ax2 = fig.add_subplot(122)

    fig.suptitle("Decomposition of %s using EGH Mixture Model Animation" % get_in_folder(), fontsize=20)
    ax1.set_title("UV Elution Decomposition", fontsize=16)
    ax2.set_title("X-ray Elution Decomposition", fontsize=16)
    ax1.plot(y1, label='input data')
    ax2.plot(y2, label='input data')

    x1 = np.arange(len(y1))
    cgys1, gy1 = get_anim_components(x1, y1, gmm1, 0)
    line1, = ax1.plot(gy1, label='eghmm-fit')
    lines1 = plot_components(ax1, x1, cgys1)

    x2 = np.arange(len(y2))
    cgys2, gy2 = get_anim_components(x2, y2, gmm2, 0)
    line2, = ax2.plot(gy2, label='eghmm-fit')
    lines2 = plot_components(ax2, x2, cgys2)

    def add_number_text(ax):
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        tx = (xmin + xmax)/2
        ty = (ymin + ymax)/2
        text = ax.text(tx, ty, "_", ha='center', va='center', alpha=0.05, fontsize=240)
        return text

    text1 = add_number_text(ax1)
    text2 = add_number_text(ax2)

    ax1.legend()
    ax2.legend()
    fig.tight_layout()
    fig.subplots_adjust(top=0.88)

    anim_artists = [line1] + lines1 + [line2] + lines2 + [text1, text2]

    def animate(i):
        # print('animate', i)

        cgys1, gy1 = get_anim_components(x1, y1, gmm1, i+1)
        line1.set_data(x1, gy1)
        for line, gy in zip(lines1, cgys1):
            line.set_data(x1, gy)

        cgys2, gy2 = get_anim_components(x2, y2, gmm2, i+1)
        line2.set_data(x2, gy2)
        for line, gy in zip(lines2, cgys2):
            line.set_data(x2, gy)

        text_num = str(i+1)
        text1.set_text(text_num)
        text2.set_text(text_num)

        dplt.update()   # need a chance to get events

        return anim_artists

    def reset():
        print('reset')
        return animate(-1)

    num_frames = gmm1.max_iter
    print('num_frames=', num_frames)
    anim = animation.FuncAnimation(fig, animate, frames=num_frames, blit=True, init_func=reset, interval=100)

    return anim

class SpikeDialog(Dialog):
    def __init__(self, parent, in_folder, **kwargs):
        self.datasets = datasets = PairedDataSets(in_folder, kwargs)
        self.continue_ = True
        self.animator = None
        self.show_extrapolated = kwargs.pop('show_extrapolated', False)
        y_list, data_list = datasets.generate_sample_datasets(quad=True)
        self.kwargs = kwargs
        Dialog.__init__(self, parent, title="SpikeDialog", visible=False)
        self.create_canvas(y_list, data_list)

    def body(self, bframe):
        self.cframe = Tk.Frame(bframe)
        self.cframe.pack()

    def buttonbox( self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Cancel", width=10, command=self._cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        w = Tk.Button(box, text="Next", width=10, command=self.draw_next)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        w = Tk.Button(box, text="Iterate", width=10, command=self.iterate)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        anim_data = self.kwargs.get('anim_data', False)
        state = Tk.NORMAL if anim_data else Tk.DISABLED
        w = Tk.Button(box, text="Animate", width=10, command=self.toggle_animation, state=state)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.toggle_btn = w

        w = Tk.Button(box, text="Step Mode", width=10, command=self.do_step)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.step_btn = w

        w = Tk.Button(box, text="Save Animation", width=16, command=self.save_animation)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.step_btn = w

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def _cancel(self):
        self.continue_ = False
        self.cancel()

    def create_canvas(self, y_list, data_list):
        cframe = self.cframe

        figsize = (20,10)
        nrows, ncols = 3, 4
        # self.fig, self.axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
        fig = plt.figure(figsize=(figsize))
        axes_list = []
        gs = GridSpec(nrows, ncols)
        for i in range(nrows):
            axes_row = []
            for j in range(ncols):
                projection = None if self.show_extrapolated or i < 2 else '3d'
                ax = fig.add_subplot(gs[i,j], projection=projection)
                axes_row.append(ax)
            axes_list.append(axes_row)

        self.fig = fig
        self.axes = np.array(axes_list)
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, cframe)
        self.toolbar.update()

        self.fig.suptitle("Decomposition of %s using Quad EGH Mixture Model" % get_in_folder(), fontsize=20)
        titles = ["UV Data at %d" % WAVELENGTH_POINTS[0], "UV Data at %d" % WAVELENGTH_POINTS[1],
                    "Xray Data at %g" % ANGLE_POINTS[0], "Xray Data at %g" % ANGLE_POINTS[1]]
        plot_hist_data_list(y_list, data_list, axes=self.axes[0,:], titles=titles)
        n_components = self.kwargs.pop('n_components', 2)
        self.mm_args = [y_list, data_list, n_components]
        self.draw_next()
        self.fig.tight_layout()
        self.fig.subplots_adjust(top=0.9)
        self.mpl_canvas.draw()

    def show(self):
        self._show()

    def draw_next(self, anim_data=False):
        self.toggle_btn.config(text="Animate")
        if self.animator is not None:
            self.animator.stop()
            self.animator = None

        axes = self.axes[1,:]
        for ax in axes:
            ax.cla()

        self.mm_list, self.xy_list = plot_decompositions(*self.mm_args, axes=axes, **self.kwargs)
        if self.show_extrapolated:
            for ano, (mm, xy, ax) in enumerate(zip(self.mm_list, self.xy_list, self.axes[2,:])):
                ax.cla()
                C = mm.get_anim_C(*xy, -1)
                self.datasets.draw_exprapolated(ano, ax, C)
        self.mpl_canvas.draw()

    def iterate(self):
        while self.continue_:
            self.draw_next()
            time.sleep(1)
            self.update()

    def create_animator(self, step_mode=False):
        from Prob.MmAnimator import MmAnimator
        axes = self.axes
        ylims = []
        for ax in axes[1,:]:
            ylims.append(ax.get_ylim())
            ax.cla()
        return MmAnimator(self.fig, axes, ylims, self.xy_list, self.mm_list, self.datasets,
                            step_mode=step_mode, show_extrapolated=self.show_extrapolated)

    def show_animation(self):
        self.animator = self.create_animator()

    def toggle_animation(self):
        text = self.toggle_btn.cget("text")
        if text == "Animate":
            self.show_animation()
            text = "Stop"
        elif text == "Stop":
            self.animator.stop()
            text = "Start"
        elif text == "Start":
            self.animator.start()
            text = "Stop"
        else:
            assert False
        self.toggle_btn.config(text=text)

    def do_step(self):
        text = self.step_btn.cget("text")
        if text == "Step Mode":
            self.start_step_mode()
            self.step_btn.config(text="Step")
        else:
            self.step_animator.step(1)
        self.mpl_canvas.draw()

    def start_step_mode(self):
        self.step_animator = self.create_animator(step_mode=True)

    def save_animation(self):
        self.animator = self.create_animator()
        self.animator.save('anim.mp4', writer="ffmpeg")
        self.animator.stop()

def spike_dialog(parent, in_folder, **kwargs):
    dialog = SpikeDialog(parent, in_folder, **kwargs)
    dialog.show()



"""

    ZeroExtrapolationOverlayBook.py

    Copyright (c) 2017-2022, SAXS Team, KEK-PF

"""
import os
import numpy                as np
import logging
# from openpyxl               import Workbook
from OpenPyXlUtil           import save_allowing_user_reply, LINE_WIDTH
from .ZeroExtrapolationResultBook import create_extrapolation_params_chart_
from molass_legacy.Reports.DefaultFont import set_default_font
set_default_font()

class ZeroExtrapolationOverlayBook:
    def __init__( self,
                wb, ws,
                mn, qvector, param_array_list, parent=None, lrf_info=None ):
        m, num_peaks_to_exec = mn

        self.logger  = logging.getLogger( __name__  )
        self.parent  = parent   # SerialController
        self.need_bq = lrf_info.need_bq()

        self.wb = wb
        self.ws = ws

        aq_asc_vector  = param_array_list[0][:,0]
        aq_desc_vector = param_array_list[1][:,0]
        bq_asc_vector  = param_array_list[0][:,1]
        bq_desc_vector = param_array_list[1][:,1]

        paren = '' if num_peaks_to_exec == 1 else '(%d)' % ( m + 1 )
        ws.title = 'Extrapolation Overlay%s' % ( paren )

        ws.append( [ 'Q', 'A(q) Asc', 'A(q) Desc', 'B(q) Asc', 'B(q) Desc' ] )

        num_rows = len(aq_asc_vector)       # <= len(qvector)
        print( 'num_rows=', num_rows )
        for i in range( num_rows ):
            a = aq_asc_vector[i]
            if a <= 0:
                a = None
            d = aq_desc_vector[i]
            if d <= 0:
                d = None
            ws.append( [ qvector[i], a, d, bq_asc_vector[i], bq_desc_vector[i] ] )

        no_orig_param_array = True
        extra_col   = 1
        col_offset  = 2

        row_offset  = 0
        j           = 0         # log scale
        param_name  = "A(q) Asc/Desc Overlayed (1.0 mg/ml)"
        aq_chart = create_extrapolation_params_chart_( ws, 2, num_rows, row_offset, col_offset, j,
                                        param_name, extra_col, no_orig_param_array, overlay=True )
        ws.add_chart( aq_chart, 'G2' )

        if self.need_bq:
            row_offset  = 0
            j           = 1         # linear scale
            param_name  = "B(q) Asc/Desc Overlayed"
            bq_chart = create_extrapolation_params_chart_( ws, 2, num_rows, row_offset, col_offset, j,
                                            param_name, extra_col, no_orig_param_array, overlay=True )
            ws.add_chart( bq_chart, 'G54' )

    def add_format_setting(self, book_file, boundary_indeces=[]):
        from .ZeroExExcelFormatter import ZeroExOverlayArgs
        args = ZeroExOverlayArgs(self.ws.title, self, book_file, boundary_indeces)  # make it picklable

        if self.parent.more_multicore:
            self.add_format_setting_more_multicore(args)
        else:
            self.add_format_setting_less_multicore(args)

    def add_format_setting_more_multicore(self, args):
        self.parent.teller.tell('range_overlay_book', args=args)

    def add_format_setting_less_multicore(self, args):
        from .ZeroExExcelFormatter import add_overlay_format_setting

        add_overlay_format_setting(self.parent.excel_client, args, self.logger)

    def save( self, xlsx_file ):
        save_allowing_user_reply( self.wb, xlsx_file )



# coding: utf-8
"""
    Rgg.LogUtils.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import logging
import numpy as np
import re
from NumpyArrayUtils import from_space_separated_list_string
from molass_legacy.Batch.StandardProcedure import StandardProcedure
from .UvAdjuster import spike_demo

def find_best_score(in_folder, log_file):
    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    sd = sp.get_corrected_sd()

    logger = logging.getLogger()

    mixture_re = re.compile(r"Mixture\((\d+)\)")
    refiner_seeds_re = re.compile(r"Refiner\((\d+), (\d+)\)")
    trucate_re = re.compile(r"^(.+\]\]), func_value=(.+)")

    parsing_params = False
    parsing_start = False
    params_key = "params="

    seeds_list = []
    params_list = []
    fv_list = []
    score_list = []
    classify_key_list = []

    mm_no = -1
    params_count = 0
    fh = open(log_file)
    for k, line in enumerate(fh.readlines()):

        if line.find("Mixture") > 0:
            m = mixture_re.search(line)
            if m:
                mm_no += 1

        if line.find("Refiner") > 0:
            parsing_params = True
            parsing_start = True
            m = refiner_seeds_re.search(line)
            if m:
                seeds = [int(m.group(n)) for n in [1,2]]
            else:
                assert False

        if parsing_params:
            # print([k], line)
            if parsing_start:
                parsing_start = False
                pos = line.find(params_key) + len(params_key)
                params_str = line[pos:]
                params_count += 1
            else:
                if line.find("]]") > 0:
                    m = trucate_re.match(line)
                    if m:
                        params_str += m.group(1)
                        fv = float(m.group(2))
                        print([mm_no], "seeds=(%d, %d), fv=%.3g" % (*seeds, fv))
                    else:
                        assert False

                    if np.isfinite(fv):
                        fv_list.append(fv)
                        parsing_params = False
                        params = from_space_separated_list_string(params_str)
                        logger.info("seeds=%s, params={%s}", str(seeds), str(params))
                        fit_score = spike_demo(in_folder, sd=sd, xr_params=params, seeds=seeds, mm_no=mm_no)
                        logger.info("fit_score=%g", fit_score)
                        fv_list.append(fv)
                        seeds_list.append(seeds)
                        params_list.append(params)
                        score_list.append(fit_score)
                        # classify_key_list.append(np.concatenate([params[:,1], [fv]]))
                        classify_key_list.append(np.concatenate([[np.std(params[:,1])/10], [fv + fit_score]]))

                else:
                    params_str += line

    fh.close()

    fv_array = np.array(fv_list)
    n = np.argmin(fv_array)
    print([n], seeds_list[n], fv_array[n])

def get_result_from_log(seeds, test_log):
    from molass_legacy.KekLib.BasicUtils import Struct

    seeds_key = "(%d, %d)" % seeds
    re_opts = re.MULTILINE | re.DOTALL
    xr_params_re = re.compile(r"xr_params=(\[\[.*?\]\])", re_opts)
    rg_params_re = re.compile(r"rg_params=(\[.*?\])", re_opts)
    tricky_params_re = re.compile(r"tricky_params=(\[.*?\])", re_opts)
    a_b_params_re = re.compile(r"\(a, b\)=(\(.*?\))", re_opts)
    uv_params_re = re.compile(r"uv_params=(\[.*?\])", re_opts)
    c_d_params_re = re.compile(r"\(c, d\)=(\(.*?\))", re_opts)
    fv_re = re.compile(r"fv=(.+)\.", re_opts)

    held_lines = []
    fh = open(test_log)
    for k, line in enumerate(fh.readlines()):
        if line.find("init_seed") > 0:
            held_lines = []
        held_lines.append(line)
        if line.find(seeds_key) > 0:
            break
    fh.close()

    log_buffer = ''.join(held_lines)
    print(log_buffer)
    m = xr_params_re.search(log_buffer)
    if m:
        xr_params = from_space_separated_list_string(m.group(1))
        print("xr_params=", xr_params)
    m = rg_params_re.search(log_buffer)
    if m:
        rg_params = from_space_separated_list_string(m.group(1))
        print("rg_params=", rg_params)
    m = tricky_params_re.search(log_buffer)
    if m:
        tricky_params = from_space_separated_list_string(m.group(1))
        print("tricky_params=", tricky_params)
        (a, b) = tricky_params
    else:
        m = a_b_params_re.search(log_buffer)
        if m:
            (a, b) = eval(m.group(1))
            print("(a, b)=", (a, b))

    m = uv_params_re.search(log_buffer)
    if m:
        uv_params = from_space_separated_list_string(m.group(1))
        print("uv_params=", uv_params)
    m = c_d_params_re.search(log_buffer)
    if m:
        (c, d) = eval(m.group(1))
        print("(c, d)=", (c, d))

    m = fv_re.search(log_buffer)
    if m:
        fv = eval(m.group(1))
        print("fv=", fv)

    n = xr_params.shape[0]
    x = np.concatenate([xr_params.flatten(), rg_params, (a, b), uv_params, (c, d)])
    return n, Struct(x=x, fun=fv)



# coding: utf-8
"""
    DMM.Spike.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import time
import numpy as np
from molass_legacy.Peaks.ElutionModels import egh
import molass_legacy.KekLib.DebugPlot as plt
from .dmm import DMM
from Prob.GaussianMixture import hist_to_source

def spike_demo():
    x = np.arange(500)
    y = np.zeros(len(x))
    for params in [
            # [0.6, 200, 40, 10],
            [0.6, 200, 30, 5],
            [1.0, 300, 30, 10],
            ]:
        y += egh(x, *params)

    spike_demo_impl(x, y, num_compoments=2)

def spike_demo_impl(x, y, num_compoments=None):
    k = 2

    dmm = DMM(k, sigma=None)
    sy = hist_to_source(x, y)
    print(dmm.estimate(sy))

    plt.push()
    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,7))
    ax1.plot(x, y)

    fig.tight_layout()
    plt.show()
    plt.pop()

def spike_demo_real(in_folder):
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    sd = sp.get_sd()
    xr_curve = sd.get_xray_curve()
    spike_demo_impl(xr_curve.x, xr_curve.y, num_compoments=5)



# coding: utf-8
"""
    CurveSaverDialog.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry

class CurveSaverDialog(Dialog):
    def __init__(self, parent, curve_list, folder_init):
        self.curve_list = curve_list
        self.folder_init = folder_init
        Dialog.__init__(self, parent, "CurveSaverDialog", visible=False)

    def body(self, frame):

        iframe = Tk.Frame(frame)
        iframe.pack(padx=40, pady=10)
        guide_frame = Tk.Label(iframe, text='Check curves and press "Save".')
        guide_frame.pack()

        space = Tk.Frame(iframe, height=20)
        space.pack()

        table_frame = Tk.Frame(iframe)
        table_frame.pack()
        curve_id = Tk.Label(table_frame, text='Curve ID', width=20)
        curve_id.grid(row=0, column=0, padx=20, sticky=Tk.W)
        length = Tk.Label(table_frame, text='Length')
        length.grid(row=0, column=1)
        filename = Tk.Label(table_frame, text='Filename')
        filename.grid(row=0, column=2)

        self.var_list = []

        for i, rec in enumerate(self.curve_list, start=1):
            cbvar = Tk.IntVar()
            cbvar.set(1)
            cb = Tk.Checkbutton(table_frame, variable=cbvar, text=rec[0])
            cb.grid(row=i, column=0, sticky=Tk.W)
            length = Tk.Label(table_frame, text='%d' % len(rec[1]))
            length.grid(row=i, column=1, sticky=Tk.E)
            filename = Tk.StringVar()
            filename.set(rec[2])
            entry = Tk.Entry(table_frame, textvariable=filename, width=30 )
            entry.grid(row=i, column=2, sticky=Tk.W, padx=20)
            self.var_list.append([cbvar, filename])

        space = Tk.Frame(iframe, height=20)
        space.pack()

        folder_frame = Tk.Frame(iframe)
        folder_frame.pack()

        folder_label = Tk.Label(folder_frame, text="Save Folder: ")
        folder_label.grid(row=0, column=0)
        self.save_folder = Tk.StringVar()
        self.save_folder.set(self.folder_init.replace('\\', '/'))
        folder_entry = FolderEntry(folder_frame, textvariable=self.save_folder, width=60)
        folder_entry.grid(row=0, column=1)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Save", width=10, command=self.save, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def show(self):
        self._show()

    def save(self):
        folder = self.save_folder.get()

        if not os.path.exists(folder):
            from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry
            mkdirs_with_retry(folder)

        for k, row in enumerate(self.var_list):
            cbvar, filename = row
            if cbvar.get() == 0:
                continue
            dir_path = os.path.join(folder, filename.get())
            np.savetxt(dir_path, self.curve_list[k][1])

        self.ok()



# coding: utf-8
"""
    QmmResultSaver.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import os
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import set_icon
from molass_legacy.KekLib.TkUtils import rational_geometry
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry, FileEntry
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry
from molass_legacy.KekLib.NumpyUtils import np_savetxt
try:
    import molass_legacy.KekLib.CustomMessageBox as MessageBox
except:
    import OurMessageBox as MessageBox

class QmmResultSaverDialog(Dialog):
    def __init__( self, parent, e11n):
        self.parent = parent
        self.e11n = e11n
        self.conc_dependence = get_setting('conc_dependence')

    def show(self):
        title = "Preview Data Saver"
        Dialog.__init__( self, self.parent, title, auto_geometry=False, geometry_cb=self.adjust_geometry )

    def adjust_geometry(self):
        rational_geometry(self, self.parent, 0.5, 0.7)

    def body(self, body_frame):
        set_icon( self )

        label_frame = Tk.Frame(body_frame)
        label_frame.pack( padx=50, pady=10 )

        guide = Tk.Label(label_frame, text='Select a folder to save into and press "OK"')
        guide.pack()

        detail_frame = Tk.Frame(body_frame)
        detail_frame.pack( padx=50, pady=10 )

        out_dir = (get_setting( 'analysis_folder' ) + '/qmm_results').replace('\\', '/')
        row = 0
        label = Tk.Label( detail_frame, text="Result Folder" )
        label.grid( row=row, column=0, sticky=Tk.W )
        self.folder = Tk.StringVar()
        self.folder.set( out_dir )
        folder_entry = FolderEntry( detail_frame, textvariable=self.folder, width=60 )
        folder_entry.grid( row=row, column=1, sticky=Tk.W, padx=10 )

    def buttonbox( self ):
        box = Tk.Frame(self)
        box.pack()
        w = Tk.Button(box, text="Save", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def validate(self, notify=True):
        try:
            self.save_data()
            if notify:
                MessageBox.showinfo("INFO", "Saved successfully.", parent=self)
            ret = 1
        except Exception as exc:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            MessageBox.showerror("ERROR", "Save failed with the following error\n" + etb.last_line(), parent=self)
            ret = 0
        return ret

    def save_data(self):
        folder = self.folder.get()
        if not os.path.exists(folder):
            mkdirs_with_retry(folder)

        e11n = self.e11n
        Cinv = e11n.Cinv
        E = e11n.E
        P = e11n.P
        v = e11n.v
        size = e11n.size
        """
        P = MC⁺
        Pe = qsrt(E*E・C⁺*C⁺)
        """
        EE = E*E
        CC = Cinv*Cinv
        print('v.shape=', v.shape)
        print('EE.shape=', EE.shape)
        print('CC.shape=', CC.shape)

        size_ = P.shape[1] if size is None else size
        Pe = np.sqrt(EE@CC)
        for j in range(size_):
            data = np.vstack( [v, P[:,j], Pe[:,j]] ).T
            np_savetxt( folder + '/A%d.dat' % j, data )



# coding: utf-8
"""

    HyperbolaFit.py

    Copyright (c) 2021, SAXS Team, KEK-PF

"""
import numpy as np
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt

USE_MINIMIZE = True
if USE_MINIMIZE:
    from scipy.optimize import minimize
else:
    from scipy.optimize import curve_fit

"""
    c.f.
    Fit a curve for data made up of two distinct regimes
"""
def hyperbola(x, a, b, c, d, e):
    """ hyperbola(x) with parameters
        a/b = asymptotic slope
         c  = curvature at vertex
         d  = offset to vertex
         e  = vertical offset
    """
    return a*np.sqrt((b*c)**2 + (x-d)**2)/b + e

def rot_hyperbola(x, a, b, c, d, e, th):
    pars = a, b, c, 0, 0 # do the shifting after rotation
    xd = x - d
    hsin = hyperbola(xd, *pars)*np.sin(th)
    xcos = xd*np.cos(th)
    return e + hyperbola(xcos - hsin, *pars)*np.cos(th) + xcos - hsin


class RotatedHyperbola:
    def __init__(self, x, y, pt=None):
        if pt is None:
            h0 = 1, 3, 0, 100, 1.5, 0.5*np.pi
        else:
            ptx, pty = pt
            h0 = pty*10/x[-1], 1, 35, ptx, pty*1.6, 0.5*np.pi
        if USE_MINIMIZE:
            self.debug = True
            def obj_func(p):
                y_ = rot_hyperbola(x, *p)
                if self.debug:
                    plt.push()
                    fig, ax = plt.subplots()
                    ax.plot(x, y)
                    ax.plot(x, y_)
                    ax.plot(ptx, pty, 'o', color='red')
                    self.debug = plt.show()
                    plt.pop()
                return np.sum((y_ - y)**2)
            result = minimize(obj_func, h0)
            h = result.x
        else:
            h, hcov = curve_fit(rot_hyperbola, x, y, h0)
        self.h = h

    def get_peak_top(self):
        ptx = self.h[3]
        pty = rot_hyperbola(ptx, *self.h)
        return ptx, pty

    def __call__(self, x):
        return rot_hyperbola(x, *self.h)

def demo():
    from .ElutionModels import gaussian

    x = np.linspace(0, 200, 201)
    y = gaussian(x, 1, 100, 30)
    slice_ = slice(70, 130)
    x_ = x[slice_]
    y_ = y[slice_]
    rh = RotatedHyperbola(x_, y_)
    print('rh.h=', rh.h)
    fig, ax = plt.subplots()
    ax.plot(x, y)
    if False:
        for rad in [1/2]:
            ax.plot(x_, rot_hyperbola(x_, 1.1, 1, 0, 100, 1.5, rad*np.pi))
    ax.plot(x_, rh(x_))
    ymin, ymax = ax.get_ylim()
    ax.set_ylim(ymin, ymax*1.1)
    plt.show()

def demo_real(in_folder):
    from matplotlib.patches import Rectangle
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Peaks.RobustPeaks import RobustPeaks
    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    sd = sp.get_corrected_sd(proxy=False)
    xr_curve = sd.get_xray_curve()
    x = xr_curve.x
    y = xr_curve.y
    rp = RobustPeaks(x, y)
    ls, pt, rs = rp.get_peaks()[1]
    print((ls, pt, rs))
    hw = (pt-ls + rs-pt)//4
    by = np.min(y[[pt-hw, pt+hw]])
    ls_ = bisect_right(y[0:pt], by)
    rs_ = pt + bisect_right(-y[pt:], -by)
    # slice_ = slice(pt-hw, pt+hw)
    slice_ = slice(ls_, rs_)
    x_ = x[slice_]
    y_ = y[slice_]

    if True:
        plt.push()
        fig, ax = plt.subplots()
        ax.plot(x, y)
        ax.plot(x_, y_)
        ax.plot(x[pt],y[pt], 'o', color='red')
        plt.show()
        plt.pop()

    try:
        rh = RotatedHyperbola(x_, y_, pt=(x[pt],y[pt]))
        print('rh.h=', rh.h)
        success = True
    except:
        success = False
    fig, ax = plt.subplots()
    ax.plot(x, y)
    if False:
        for rad in [1/2]:
            ax.plot(x_, rot_hyperbola(x_, 1.1, 1, 0, 100, 1.5, rad*np.pi))

    if success:
        ax.plot(x_, rh(x_))
        ptx, pty = rh.get_peak_top()
        ax.plot(ptx, pty, 'o', color='red')

    ymin, ymax = ax.get_ylim()
    ymax_ = ymax*1.1
    ax.set_ylim(ymin, ymax_)

    f, t = x[[ls,rs]]
    p = Rectangle(
            (f, ymin),      # (x,y)
            t - f,          # width
            ymax_ - ymin,    # height
            facecolor   = 'cyan',
            alpha       = 0.2,
        )
    ax.add_patch(p)

    plt.show()



# coding: utf-8
"""
    ClusteringUtils.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import time
import numpy as np
import scipy.stats as stats
from scipy.optimize import fsolve
from sklearn.cluster import AgglomerativeClustering, KMeans
import molass_legacy.KekLib.DebugPlot as plt
from .ProbDataUtils import generate_samle_datasets

def spike_demo(in_folder, **kwargs):
    n_components = kwargs.pop('n_components', 2)

    y1, y2, data1, data2 = generate_samle_datasets(in_folder, kwargs)

    st = time.time()
    # clustering = AgglomerativeClustering(n_clusters=n_components, linkage='ward').fit(X = np.expand_dims(data1,1))
    clustering = KMeans(n_clusters=n_components).fit(X = np.expand_dims(data1,1))
    elapsed_time = time.time() - st
    labels = clustering.labels_
    print("Elapsed time: %.2fs" % elapsed_time)
    print("Number of points: %i" % labels.size)
    print('labels[:5]=', labels[:5])
    label_values = np.unique(labels)
    print('label_values', label_values)

    num_bins = len(y1)
    print('num_bins=', num_bins)
    plt.push()
    fig  = plt.figure()
    ax = fig.gca()
    ax.hist(data1, bins=num_bins)
    ax.set_ylim(ax.get_ylim())

    for k in label_values:
        data = data1.copy()
        data[labels!=k] = 0
        hist, bin_edges = np.histogram(data, bins=num_bins, range=(0, num_bins-1))
        print([k], 'len(hist)=', len(hist))
        hist[0] = 0
        ax.plot(hist)

    fig.tight_layout()

    plt.show()
    plt.pop()



"""
    Models.RateTheory.EdmBasicStudy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from molass_legacy.Models.RateTheory.RobustEDM_Utils import compute_moments
from molass_legacy.Models.Stochastic.DispersivePdf import dispersive_monopore_pdf
from .DispersivePdf import N0

def guess_sdm_params(x, y, timescale=0.25):
    M = compute_moments(x, y)
    print("M=", M)

    t0_init = 50
    rho = 0.5
    me = 1.5
    mp = 1.5

    def moment_objective(p):
        K, N, t0  = p
        T = K/N
        n1 = N*(1 - rho)**me
        t1 = T*(1 - rho)**mp
        M1_ = t0 + n1*t1
        M2_ = 2*n1*t1**2 + M1_**2/N0
        M3_ = 6*n1*t1**2*(N0*t1 + n1*t1 + t0)/N0
        return np.log((M1_ - M[1])**2) + np.log((M2_ - M[2])**2) + np.log((M3_ - M[3])**2)
    
    res = minimize(moment_objective, [500, 1000, t0_init], method='Nelder-Mead')

    K_init, N_init, t0_init2 = res.x

    def scale_objective(p):
        K, N, t0, scale  = p
        T = K/N
        n1 = N*(1 - rho)**me
        t1 = T*(1 - rho)**mp
        y_ = scale*dispersive_monopore_pdf(x, n1, t1, N0, t0, timescale=timescale)
        return np.sum((y_ - y)**2)

    res = minimize(scale_objective, [K_init, N_init, t0_init2, M[0]], method='Nelder-Mead')
    K, N, t0, scale = res.x

    return np.array([K, N, t0, N0, rho, me, mp, scale])


# coding: utf-8
"""
    LmfitThreadSafe.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""
import threading
import lmfit
from lmfit                  import Parameters

lock = threading.Lock()

def minimize( *args, **kwargs ):
    with lock:
        result = lmfit.minimize( *args, **kwargs )
    return result

if __name__ == '__main__':
    import numpy    as np
    import time
    from nose.tools import eq_

    def func_threaded( result_list ):
        params = Parameters()
        params.add( 'A', value=0,       min=-1,  max=+1 )

        def obj_func( p ):
            A = p['A']
            array = np.array( [ A - 0.5 ] )
            time.sleep( 0.1 )
            return array

        result = minimize( obj_func, params, args=() )
        print( "result.params['A']=", result.params['A'].value )
        result_list.append( result.params['A'].value )

    thread_list = []
    result_list = []
    for i in range(3):
        thread = threading.Thread(
                    target=func_threaded,
                    name='Thread%d' % i,
                    args=[ result_list ],
                    )
        thread.start()
        thread_list.append( thread )

    for thread in thread_list:
        thread.join()

    print( 'result_list=', result_list )
    eq_( result_list, [ 0.5, 0.5, 0.5 ] )



"""
    Processes.GUI.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
def gui_loop(si, pn, queues):
    print("gui_loop")

    from molass_legacy.KekLib.TkUtils import get_tk_root

    root = get_tk_root()

    root.after(100, lambda: gui_main(si, pn, root, queues))
    root.mainloop()
    root.destroy()

def gui_main(si, pn, root, queues):
    from GuiProcess.GuiMain import GuiMain
    print("gui_main", si)

    gm = GuiMain(root, si, pn, queues)
    gm.show()

    for q, stop_signal in zip(queues[1:], ["__stop__", -9, -9]):
        q.put(stop_signal)

    root.quit()
    print("gui quit")



"""
    Trimming.__init__.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from .CdLimit import CdLimit
from .FlangeLimit import FlangeLimit
from .GuinierLimit import GuinierLimit
from .FlowChange import FlowChange
try:
    from .DataRange import DataRangeDialog
except ImportError:
    pass
from .TrimmingInfo import *
from .AutoRestrictor import AutoRestrictor



"""
    OptLrfInfoDebug.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Optimizer.OptLrfInfo import OptLrfInfo

class OptLrfInfoProxy(OptLrfInfo):
    def __init__(self,
                    qv, xrD, xrE,
                    x, y, xr_ty, xr_cy_list,
                    uv_x, uv_y, uv_ty, uv_cy_list,
                    composite):
        self.matrices = (None,) * 5
        self.qv = qv
        self.xrD = xrD
        self.xrE = xrE
        self.x = x
        self.y = y
        self.xr_ty = xr_ty
        self.scaled_xr_cy_array = np.asarray(xr_cy_list)
        self.uv_x = uv_x
        self.uv_y = uv_y
        self.uv_ty = uv_ty
        self.scaled_uv_cy_array = np.asarray(uv_cy_list)
        self.composite = composite


"""
    GuinierExcelFormatter.py

    Copyright (c) 2016-2025, SAXS Team, KEK-PF
"""
import os
import numpy as np
from molass_legacy.KekLib.NumpyUtils import get_safe_approximate_max
from molass_legacy.KekLib.ExcelCOM import xlValue, xlLine, compute_axis_max_value
from .DefaultFont import set_default_bookfont_com

NUM_CHARTS = 5
CHART_XFIX_300  = 14.5
CHART_AREA_LEFT = 816.0     # chart_area.Left of the basis environment
NUM_POINTS = 160            # num_points of the basis environment
ADJUST_PLOT_LEFT = False
TEXT_Y_ADJUST_UNIT = 20
TEXT_HALF_WIDTH = 120

def chart_xfix(xoffset, n, adjust_scale):
    return xoffset + CHART_XFIX_300 * n/300 * adjust_scale

class GuinierReportArgs:
    def __init__(self, title, book_obj, book_file, ranges):
        self.title = title
        self.num_points = book_obj.num_points
        self.j0 = book_obj.j0
        self.Izc_array = book_obj.Izc_array
        self.include_end_consistency = book_obj.include_end_consistency
        self.iz_array = book_obj.iz_array
        self.rg_array = book_obj.rg_array
        self.book_file = book_file
        self.ranges = ranges

def get_analysis_area(args, ole_ws, interval, logger, debug=False):

    if ADJUST_PLOT_LEFT:
        left_poss = []
        for k in range(NUM_CHARTS):
            chart = ole_ws.get_chart(k)
            left_poss.append(chart.PlotArea.Left)
        average_plot_left = np.average(left_poss, axis=0)

    top_chart = ole_ws.get_chart( 0 )
    top_area  = top_chart.ChartArea
    conc_chart = ole_ws.get_chart( 0 )
    chart_area = conc_chart.ChartArea
    plot_area = conc_chart.PlotArea

    adjust_scale = chart_area.Left/CHART_AREA_LEFT * NUM_POINTS/args.num_points
    # xoffset = plot_area.Width/args.num_points * 0.5
    xoffset = 0

    if debug:
        logger.info("get_analysis_area: j0=%d, num_points=%d, inverval=%s", args.j0, args.num_points, str(interval))
        if ADJUST_PLOT_LEFT:
            logger.info("get_analysis_area: average_plot_left = %s", str(average_plot_left))
        logger.info("get_analysis_area: (chart_area.Left, plot_area.Left) = %s", str((chart_area.Left, plot_area.Left)))
        logger.info("get_analysis_area: (chart_area.Width, plot_area.Width) = %s", str((chart_area.Width, plot_area.Width)))

    x = []
    for j in interval:
        point = ole_ws.get_chart_data_point( conc_chart, 0, j )
        xpos = chart_xfix(xoffset, j, adjust_scale)
        x.append(point.Left + xpos)
        if debug:
            logger.info("%s (point.Left, point.Width, xpos) = %s", str([j]), str((point.Left, point.Width, xpos)))

    bottom_chart = ole_ws.get_chart( ole_ws.get_num_charts() - 1 )
    bottom_area = bottom_chart.ChartArea

    if debug:
        logger.info("(bottom_area.Top, bottom_area.Height) = %s", str((bottom_area.Top, bottom_area.Height)))

    if ADJUST_PLOT_LEFT:
        plot_area_adjust = average_plot_left - plot_area.Left
    else:
        plot_area_adjust = 0
    chart_left = chart_area.Left + plot_area_adjust
    x_left  = chart_left + x[0]
    x_right = chart_left + x[1]
    y_top   =    0
    y_btm   = top_area.Top + bottom_area.Top + bottom_area.Height

    return [ x_left, x_right, y_top, y_btm ]

def add_guinier_annonations(excel_client, args, logger, debug=False):
    # print( 'add_annonations: ranges=', ranges )

    book_file = args.book_file
    ranges = args.ranges

    book_path = os.path.abspath( book_file )
    sheet_path = book_path + '(' + args.title  + ')'
    if debug:
        print("add_guinier_annonations: sheet_path=", sheet_path)
        print("add_guinier_annonations: ranges=", str(ranges))

    ws = excel_client.openWorksheet( sheet_path )
    set_default_bookfont_com(ws.workbook)

    izc_max_axis = None
    range_index_list = []
    last_x_center = None
    for k, interval in enumerate(ranges):
        x_left, x_right, y_top, y_btm = get_analysis_area(args, ws, interval, logger, debug=debug)

        if debug:
            logger.info("add_guinier_annonations: %s interval=%s", str([k]), str(interval))
            logger.info("(x_left, x_right, y_top, y_btm) = %s", str((x_left, x_right, y_top, y_btm)))

        range_index_list += list( range( interval[0], interval[1]+1 ) )

        x_center = ( x_left + x_right ) / 2

        if last_x_center is None or last_x_center + TEXT_HALF_WIDTH < x_center - TEXT_HALF_WIDTH:
            y_btm_adjust = 0
        else:
            y_btm_adjust = TEXT_Y_ADJUST_UNIT

        last_x_center = x_center

        y_btm += y_btm_adjust
        line1 = ws.draw_line( ( x_left,  y_top ), ( x_left,  y_btm ) )
        line2 = ws.draw_line( ( x_right, y_top ), ( x_right, y_btm ) )
        b_width = x_right - x_left
        x_ = x_left + b_width//2 - 10
        y_ = y_btm - b_width//2 + 12
        try:
            brace = ws.draw_right_brace( ( x_, y_ ), 20, b_width, 90 )
        except Exception as exc:
            logger.warning( str(exc) )
            continue

        for i, shape in enumerate( [ line1, line2, brace ] ):
            shape.Select()
            line_ = excel_client.selection().ShapeRange.Line
            line_.ForeColor.RGB =   0x50B000    # RGB(0, 176, 80)
            line_.Weight = 1

        interval_repo = [args.j0 + r for r in interval]
        ws.draw_textbox( "Concentration Dependency Analysis Interval %s" % str( interval_repo ), ( x_center - TEXT_HALF_WIDTH, y_btm + 30 ), 300, 20, visible=False )

        slice_ = slice( interval[0], interval[1]+1 )
        izc_array = args.Izc_array[slice_]
        # Be careful for inf values which make Excel hang!
        try:
            izc_max_axis_ =  np.percentile( izc_array[ np.isfinite(izc_array) ], 95 ) * 2
            if izc_max_axis is None or izc_max_axis_ > izc_max_axis:
                izc_max_axis = izc_max_axis_
        except:
            # i.e., in case of no finite values?
            # occured in 20151119/Kosugi3a_Backsub
            pass

    if izc_max_axis is None:
        # i.e., in case of no range
        izc_max_axis =  np.percentile( args.Izc_array[ np.isfinite(args.Izc_array) ], 95 ) * 2

    if len(range_index_list) > 0:
        range_index = np.array( range_index_list )
    else:
        range_index = slice(0,None)

    # modify QualityChar
    # print( 'modify QualityChar' )
    quality_chart = ws.get_chart( 1 )
    seriesno = 6 if args.include_end_consistency else 5
    series = quality_chart.FullSeriesCollection(seriesno)   # atsas quality
    series.ChartType = xlLine
    series.Format.Line.Weight = 1.5
    quality_chart.Axes(xlValue).MaximumScale = 1

    # modify IzChart
    # print( 'modify IzChart' )
    iz_max = get_safe_approximate_max( args.iz_array[range_index] )
    # iz_max_axis = float( '%.2g' % ( iz_max * 1.2 ) )
    iz_max_axis = compute_axis_max_value( iz_max * 1.2 )
    # print( 'iz_max=', iz_max, 'iz_max_axis=', iz_max_axis )

    iz_chart = ws.get_chart( 2 )
    iz_chart.Axes(xlValue).MaximumScale = iz_max_axis

    # modify z_devided_by_C_Chart
    # print( 'izc_max_axis=', izc_max_axis )
    izc_chart = ws.get_chart( 3 )
    izc_chart.Axes(xlValue).MinimumScale = 0
    izc_chart.Axes(xlValue).MaximumScale = izc_max_axis

    # modify RgChart
    # print( 'modify RgChart' )
    rg_primary_max = get_safe_approximate_max( args.rg_array[range_index] )

    # print("args.rg_array=", args.rg_array)
    args.rg_array[np.isnan(args.rg_array)] = 0  # temporary fix

    hc, hb = np.histogram( args.rg_array, bins=10 )
    # hb_ = [ ( hb[i]+hb[i+1] )/2  for i in range(len(hc)) ]
    rg_hist = sorted( zip( hc, hb[1:] ), reverse=True )
    rg_hist_max = np.max( rg_hist[:3], axis=0 )[1]

    rg_max = min( rg_primary_max * 2.0, max( rg_primary_max * 1.3, rg_hist_max ) )
    rg_max_axis = compute_axis_max_value( rg_max )
    # print( 'rg_primary_max=', rg_primary_max, 'rg_hist_max=', rg_hist_max, 'rg_max_axis=', rg_max_axis )

    rg_chart = ws.get_chart( 4 )
    rg_chart.Axes(xlValue).MaximumScale = rg_max_axis

    # Zoom and Scroll
    # print( 'Zoom and Scroll' )
    ws.zoom( 70 )
    scroll_column = 17 if args.include_end_consistency else 20
    ws.freeze_panes( "D2", scroll_column=scroll_column )

    ws.workbook.Save()
    ws.workbook.Close()     # required in the more_multicore mode



"""
    Optimizer.OptLogFile.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import re

class OptLogFile:
    def __init__(self, path):
        from molass_legacy.Optimizer.StrategicOptimizer import LOG_IDENTIFICATION

        version_re = re.compile(r"\((\d+-\d+-\d+)")
        optlist_re = re.compile(r"optlist=(\[[^\]]+\])")
        active_indices_re = re.compile(r"indeces=(\[[^\]]+\])")

        self.active_indeces = None
        optlist = None
        version_date = None
        
        with open(path) as fh:
            for k, line in enumerate(fh.readlines()):
                if line.find("_MOLASS") > 0:
                    m = version_re.search(line)
                    if m:
                        version_date = m.group(1)
                    continue

                if line.find("started with") > 0:
                    m = optlist_re.search(line)
                    if m:
                        optlist = eval(m.group(1))
                    continue

                if line.find(LOG_IDENTIFICATION) > 0:
                    # note that use of Strategic Optimizer is optional
                    m = active_indices_re.search(line)
                    if m:
                        self.active_indeces = eval(m.group(1))
                    break

        assert optlist is not None

        self.version_date = "2022-09-16" if version_date is None else version_date
        self.optdict = dict(optlist)
        self.class_code = self.optdict['-c']
        self.nc = int(self.optdict['-n'])

    def get_active_indeces(self):
        if self.active_indeces is None:
            self.active_indeces = []
        return self.active_indeces


"""
    SecTheory.ShapeRgFiguree.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

def plot_ellipsoid(ax, a, b, c, max_radius=None):
    coefs = (a, b, c)  # Coefficients in a0/c x**2 + a1/c y**2 + a2/c z**2 = 1 
    # Radii corresponding to the coefficients:
    rx, ry, rz = 1/np.sqrt(coefs)

    # Set of all spherical angles:
    u = np.linspace(0, 2 * np.pi, 20)
    v = np.linspace(0, np.pi, 20)

    # Cartesian coordinates that correspond to the spherical angles:
    # (this is the equation of an ellipsoid):
    x = rx * np.outer(np.cos(u), np.sin(v))
    y = ry * np.outer(np.sin(u), np.sin(v))
    z = rz * np.outer(np.ones_like(u), np.cos(v))

    # Plot:
    # ax.plot_surface(x, y, z,  rstride=4, cstride=4, color='b')
    ax.plot_surface(x, y, z, alpha=0.5)
    # ax.plot_wireframe(x, y, z)

    # Adjustment of the axes, so that they all have the same span:
    if max_radius is None:
        max_radius = max(rx, ry, rz)
    for axis in 'xyz':
        getattr(ax, 'set_{}lim'.format(axis))((-max_radius, max_radius))

def compute_weights(a, b, c):
    s = np.power(a*b*c, -1/3)
    return a*s, b*s, c*s

def compute_Rg(a, b, c):
    Rg = np.sqrt((a**2 + b**2 + c**2)/5)
    return Rg
        
def plot_ellipsoid_rg_ratio(fig):

    
    axes = []
    for n in [141, 142, 143]:
        ax = fig.add_subplot(n, projection='3d')
        ax.set_axis_off()
        axes.append(ax)
    ax2 = fig.add_subplot(144)
   
    for k, (a, b, c) in enumerate([(0.1,1,1), (1,1,1), (10,1,1)]):
        ax = axes[k]
        a_, b_, c_ = compute_weights(a, b, c)
        Rg = compute_Rg(a_, b_, c_)
        ax.set_title("a=%.2g, b=%.2g, c=%.2g; $R_g$=%.2g" % (a_, b_, c_, Rg))
        plot_ellipsoid(ax, a_, b_, c_, max_radius=2.5)
    av = np.linspace(0.1, 10, 40)
    av_, bv_, cv_ = compute_weights(av, b, c)
    Rg = compute_Rg(av_, bv_, cv_)
    ax2.set_title("$R_g$ vs. a in (0.1, 10)")
    ax2.plot(av, Rg)
    ax2.set_xlabel("a")
    ax2.set_ylabel("$R_g$")

if __name__ == '__main__':
    import seaborn
    seaborn.set()

    fig = plt.figure(figsize=(16,4))
    fig.suptitle("Rg dependence on Shapes with the same Volume as a unity radius Sphere")
    plot_ellipsoid_rg_ratio(fig)
    fig.tight_layout()
    plt.show()



"""
    SecTheory.MwRgFigure.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from .MwRgInfo import get_mwrg_info
import molass_legacy.KekLib.CustomMessageBox as MessageBox

class MwRgFigure(Dialog):
    def __init__(self, parent, excl_limit, **kwargs):
        self.excl_limit = excl_limit
        self.pickable = kwargs.pop('pickable', True)
        Dialog.__init__(self, parent, "Mw-Rg Figure", visible=False, **kwargs)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()

        self.fig, self.ax = plt.subplots()
        self.draw_mwrg_figure(self.ax)
        self.fig.tight_layout()

        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.draw()

        if self.pickable:
            self.mpl_canvas.mpl_connect('button_press_event', self.on_click)

    def buttonbox(self):
        lower_frame = Tk.Frame(self)
        lower_frame.pack(fill=Tk.BOTH, expand=1)

        width = int(self.mpl_canvas_widget.cget('width'))
        padx = width*0.05

        tframe = Tk.Frame(lower_frame)
        tframe.pack(side=Tk.LEFT, padx=padx)
        self.toolbar = NavigationToolbar( self.mpl_canvas, tframe )
        self.toolbar.update()

        space = Tk.Frame(lower_frame, width=padx)
        space.pack(side=Tk.RIGHT)

        box = Tk.Frame(lower_frame)
        box.pack(side=Tk.RIGHT)

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=padx, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def draw_mwrg_figure(self, ax):
        mwrg_info = get_mwrg_info(devel=True)
        self.mwrg_info = mwrg_info

        excl_limit =  self.excl_limit
        Rg = mwrg_info.compute_rg(excl_limit)

        ax.set_title(r"Rg=%.3g ($\AA$) from the Exclusion Limit %g (kDa)" % (Rg, excl_limit), fontsize=20)
        ax.set_xlabel("Molecular Weight (kDa)", fontsize=16)
        ax.set_ylabel(r"Rg ($\AA$)", fontsize=16)

        ax.plot(mwrg_info.mws, mwrg_info.rgs, "o")

        ax.plot(excl_limit, Rg, "o", color="red", alpha=0.2, markersize=20)

        # xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        xmax = excl_limit * 1.3
        ymax = Rg * 1.3
        xmin = -xmax*0.1
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)

        x = np.linspace(xmin, xmax, 100)
        lower = mwrg_info.compute_rg(x, n_sigma=-1)
        upper = mwrg_info.compute_rg(x, n_sigma=+1)
        ax.fill_between(x, lower, upper, color="cyan", alpha=0.2, label=r"1$\sigma$ uncertainty")

        y = mwrg_info.compute_rg(x)
        ax.plot(x, y, ":")

        lower_rg = mwrg_info.compute_rg(excl_limit, n_sigma=-1)
        upper_rg = mwrg_info.compute_rg(excl_limit, n_sigma=+1)
        ax.plot([xmin, excl_limit], [Rg, Rg], ":", color="gray")
        ax.plot([excl_limit, excl_limit], [ymin, upper_rg], ":", color="gray")

        for rg in [lower_rg, upper_rg]:
            ax.plot([xmin, excl_limit], [rg, rg], ":", color="gray")

        dx = (xmax - xmin)*0.05
        dy = (ymax - ymin)*0.05
        for rg in [lower_rg, Rg, upper_rg]:
            ax.annotate("%.3g" % rg, xy=(xmin, rg), xytext=(xmin + dx, rg - dy),
                        ha='left', va="top", arrowprops=dict(arrowstyle="->", color='gray'),
                        fontsize=16)

        ax.legend(loc="lower right", fontsize=16)

    def on_click(self, event):
        if event.button != 1:
            return

        if event.xdata is None or event.ydata is None:
            return

        mws = self.mwrg_info.mws
        rgs = self.mwrg_info.rgs
        dist2 = (mws - event.xdata)**2 + (rgs - event.ydata)**2
        m = np.argmin(dist2)
        # print("m=", m, self.mwrg_info.names[m])
        MessageBox.showinfo("Mw-Rg Data Info",
            "Mw=%.3g, Rg=%.3g for %s" % (mws[m], rgs[m], self.mwrg_info.names[m]),
            parent=self
            )



"""
    NpSharedMemory.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from multiprocessing import shared_memory
from molass_legacy._MOLASS.SerialSettings import get_setting

class NpSharedMemory:
    def __init__(self, dtype=int, size=2, name=None):
        if name is None:
            a = np.zeros(size, dtype=dtype)
            self.shm = shared_memory.SharedMemory(create=True, size=a.nbytes)
            self.name = self.shm.name
            self.array = np.ndarray(a.shape, dtype=dtype, buffer=self.shm.buf)
            self.array[:] = a[:]
        else:
            self.shm = shared_memory.SharedMemory(name=name)
            self.name = name
            self.array = np.ndarray((size,), dtype=dtype, buffer=self.shm.buf)

    def close(self):
        if self.shm is not None:
            self.shm.close()
            self.shm = None

    def unlink(self):
        if self.shm is not None:
            print("--------------- unlink")
            self.shm.close()
            self.shm.unlink()
            self.shm = None

    def __del__(self):
        self.close()

shm_singleton = None

def create_shm_singleton():
    global shm_singleton
    using_shared_memory = get_setting("using_shared_memory")
    if using_shared_memory:
        shm_singleton = NpSharedMemory()

def destroy_shm_singleton():
    global shm_singleton
    if shm_singleton is not None:
        shm_singleton.unlink()
        shm_singleton = None

def get_shm_singleton():
    if shm_singleton is None:
        create_shm_singleton()
    return shm_singleton

def get_shm_proxy(name):
    return NpSharedMemory(name=name)

def close_shared_memory():
    global shm_singleton
    if shm_singleton is not None:
        shm_singleton.unlink()
        shm_singleton = None



# coding: utf-8
"""

    ElutionBaseCurve.py

    scattering baseline solver assuming Gaussian Curves

    Copyright (c) 2017-2020, SAXS Team, KEK-PF

"""
import numpy                as np
from scipy                  import stats
from bisect                 import bisect_right
# from scipy.interpolate      import UnivariateSpline
from scipy.interpolate      import LSQUnivariateSpline
import logging
from BasePercentileOffset   import base_percentile_offset, DEFAULT_OFFSET
from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker

DEBUG   = False

FWHM_SIGMA_RATIO        = 2*np.sqrt(2*np.log(2))
ACCEPTABLE_WIDTH_MIN    = 10
ACCEPTABLE_RATIO_LOWS   = 0.4
DEFAULT_SIZE_SIGMA      = 5
NUM_REGRESSION_POINTS   = 20

def get_spline_smoothing( max_y ):
    adjust = -0.5 if max_y > 0.01 else 0.5
    smoothing = np.power( 10, np.log( max_y ) + adjust )
    return smoothing

def get_spline_smoothing_better( y, max_y ):
    x   = np.arange(NUM_REGRESSION_POINTS)
    slope, intercept, r_value, p_value, std_err = stats.linregress( x, y[0:NUM_REGRESSION_POINTS] )
    s   = std_err**2 * 1e3 * len( y )
    # print( 'std_err=', std_err )
    # print( 'max_y=', max_y, 's=%.1e' % s )
    return s

class ElutionBaseCurve:
    def __init__( self, y ):
        self.y      = y
        self.max_n  = np.argmax( y )
        # self.max_y  = y[self.max_n] - np.min( y )
        self.max_y  = y[self.max_n]
        self.x      = np.arange( len(self.y) )
        self.sorted_y = sorted( y )
        self.height = self.sorted_y[-1] - self.sorted_y[0]
        self.noisiness = None
        self.logger = logging.getLogger( __name__ )

    def compute_noisiness( self ):
        if self.noisiness is None:
            # smoothing = get_spline_smoothing_better( self.y, self.max_y )
            # spline = UnivariateSpline( self.x, self.y, s=smoothing )
            knots = np.linspace( 0, len(self.y), len(self.y)//10 )
            self.spline = LSQUnivariateSpline( self.x, self.y, knots[1:-1] )
            self.ys = self.spline( self.x )
            std_diff = np.std( self.y - self.ys )
            self.noisiness = noisiness = std_diff/self.height

        return self.noisiness

    def recognize_peaks( self ):
        print('recognize_num_peaks')
        ylow, yhigh = np.percentile( self.y, [ 30, 70 ] )
        ipp = self.y < ylow
        x_  = self.x[ipp]
        y_  = self.y[ipp]
        slope, intercept, r_value, p_value, std_err = stats.linregress( x_, y_ )
        line = slope*self.x + intercept
        y   = self.y - line
        ipp = self.y > yhigh
        # print( 'ipp=', ipp )
        peaks = []
        last_b = False
        f = None
        # TODO: faster algorithm;
        for i, b in enumerate( ipp ):
            if b:
                if last_b:
                    continue
                else:
                    f = i
            else:
                if last_b:
                    peaks.append( [ f, i-1 ] )
                else:
                    f = i
            last_b = b

        if last_b:
            peaks.append( [ f, i-1 ] )

        # print( 'peaks=', peaks )
        connected_peaks = []
        last_p = None
        for p in peaks:
            if last_p is not None:
                if p[0] - last_p[1] > 2:
                    connected_peaks.append( last_p )
                else:
                    last_p[1] = p[1]
                    continue

            last_p = p

        connected_peaks.append( last_p )

        # print( 'connected_peaks=', connected_peaks )
        if False:
            from DebugCanvas import DebugCanvas
            def debug_plot( fig ):
                ax = fig.add_subplot( 111 )
                ax.set_title( 'Debug Plot' )
                ax.plot( self.y )

                for p in connected_peaks:
                    slice_ = slice( p[0], p[1]+1 )
                    ax.plot( self.x[slice_], self.y[slice_], color='red' )

            dc = DebugCanvas( "Debug Plot", debug_plot )
            dc.show()

        return connected_peaks

    def split( self, peaks ):
        assert len( peaks ) > 1

        curves = []
        for i, p in enumerate(peaks[1:]):
            prev = peaks[i]
            boundary = ( prev[1] + p[0] )//2
            curve = ElutionBaseCurve( self.y[prev[0]:boundary] )
            curves.append( curve )
        curve = ElutionBaseCurve( self.y[boundary:] )
        curves.append( curve )
        return curves

    def compute_peak_width( self ):
        peaks = self.recognize_peaks()
        if len( peaks ) > 1:
            curves = self.split( peaks )
            widths = []
            for curve in curves:
                w = curve.compute_peak_width_impl()
                # print( 'w=', w )
                if w is not None:
                    widths.append( w )

            # TODO: consider asymmetry
            return np.max( widths ) * len( widths )
        else:
            return self.compute_peak_width_impl()

    def compute_peak_width_impl( self ):
        hm  = self.max_y / 2
        y_L = self.y[0:self.max_n]
        i_L = bisect_right( sorted( y_L ), hm )
        y_R = self.y[self.max_n:]
        j   = bisect_right( sorted( y_R ), hm )
        i_R = len(self.y) - j
        self.i_L    = i_L
        self.i_R    = i_R
        self.width  = i_R - i_L
        self.ret_width  = None

        if self.width < ACCEPTABLE_WIDTH_MIN or i_R == len(self.y):
            return self.ret_width

        hm  = np.min( self.y[[i_L, i_R]] )
        num_lows = np.sum( self.y[i_L:i_R] < hm )
        self.ratio  = num_lows / self.width
        if self.ratio > ACCEPTABLE_RATIO_LOWS:
            return self.ret_width

        self.ret_width  = self.width
        return self.ret_width

    def compute_size_sigma( self ):
        try:
            peak_width  = self.compute_peak_width()
            print( 'peak_width=', peak_width )
            if peak_width is None:
                size_sigma  = DEFAULT_SIZE_SIGMA
            else:
                size_sigma  = len(self.y) * 0.5 / peak_width * FWHM_SIGMA_RATIO
        except:
            size_sigma = 5
            etb = ExceptionTracebacker()
            self.logger.warning( str(etb) + '; size_sigma is set to ' + str(size_sigma) )

        if DEBUG:
            print( 'n=', self.max_n )
            print( 'peak_width=', peak_width  )
            print( 'size_sigma=', size_sigma )

        return size_sigma

    def compute_base_percentile_offset( self, q, size_sigma=None, return_tuple=False, return_simple=False ):
        # See Also BasePercentileOffset.py
        # return DEFAULT_OFFSET

        noisiness   = self.compute_noisiness()
        if return_simple:
            bpo_simple = bisect_right( self.sorted_y, self.height*noisiness )/len( self.y )*30  # 30 = 100 * 0.3
            return bpo_simple

        if size_sigma is None:
            size_sigma  = self.compute_size_sigma()
        bpo = base_percentile_offset( q, noisiness, size_sigma=size_sigma )
        if return_tuple:
            return bpo, noisiness, size_sigma
        else:
            return bpo



"""

    ElutionMapperCanvas.py

        recognition of peaks

    Copyright (c) 2018-2023, SAXS Team, KEK-PF

"""
import os
import numpy as np
import copy
import time
import threading
import logging
import queue
from scipy.interpolate import UnivariateSpline
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
import molass_legacy.KekLib.CustomMessageBox as MessageBox
from molass_legacy.KekLib.TkUtils import is_low_resolution
from molass_legacy.KekLib.BasicUtils import get_caller_module
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable, BlinkingFrame
from molass_legacy.QuickAnalysis.AnalysisGuide import get_analysis_guide_info
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting, get_xray_picking
from molass_legacy.UV.PlainCurveUtils import get_flat_wavelength
from molass_legacy.SerialAnalyzer.AnalyzerUtil import get_init_ranges
from molass_legacy.DataStructure.AnalysisRangeInfo import set_default_analysis_range_info
from .ElutionMapper import ElutionMapper
# from .ElutionMapperPlotter import ElutionMapperPlotter, SUGGEST_DISCOMFORT_BY_TEXT
from .ElutionMapperAdjuster import ElutionMapperAdjuster
from .MappingParams import get_mapper_opt_params
from molass_legacy._MOLASS.Version import is_developing_version

show_button_text_dict = {   'locally'   : 'Show Locally Scaled Curves', 
                            'uniformly' : 'Show Uniformly Scaled Curves',
                            }

DRAW_MAPPING_RANGES     = True
USE_NEW_3D_PLOT_CLASS   = True
ENABLE_SIMULATION       = False

class ElutionMapperCanvas(Dialog):
    def __init__(self, dialog, serial_data, sd_orig, pre_recog_orig, mapper, judge_holder, initial_navigation=True):
        assert serial_data.pre_recog is not None

        self.logger         = logging.getLogger( __name__ )
        self.grab = 'local'     # used in grab_set
        self.dialog = dialog
        self.parent = dialog.parent
        self.applied        = False
        self.ok_debugging   = False
        self.caller_module  = get_caller_module( level=2 )
        self.developer_mode = False
        self.absorbance_picking = get_setting('absorbance_picking')
        self.intensity_picking  = get_xray_picking()
        self.logger.info("dialog will be shown in the mapping plane with λ=%g and q=%g", self.absorbance_picking, self.intensity_picking)
        self.logger.info("UV extinct wavelength λ=%g", get_flat_wavelength(serial_data.lvector))

        self.serial_data    = serial_data

        self.sd_orig = sd_orig                  # currently used only for do_devel_test in QuickAnalysis.PreDecomposer.py
        self.pre_recog_orig = pre_recog_orig    # currently used only for do_devel_test in QuickAnalysis.PreDecomposer.py

        self.a_vector_size  = serial_data.conc_array.shape[1]       # better than len(self.mapper.a_vector) ?
        self.qvector        = serial_data.qvector
        self.intensity_array    = serial_data.intensity_array
        self.xray_slice     = serial_data.xray_slice
        self.mapper = mapper
        self.judge_holder = judge_holder
        self.judge_holder.set_mapper(mapper)

        self.update_range_info()

        self.mapping_show_mode  = 'locally'
        self.initial_std_diff   = self.mapper.std_diff
        self.current_std_diff   = self.mapper.std_diff
        self.final_std_diff     = None
        self.initial_navigation = initial_navigation
        self.state_depedents    = []
        self.get_guide_info()
        Dialog.__init__( self, self.parent, "Elution Mapping Confirmation", visible=False)
        self.update_button_colors()
        self.decomp_editor = None

    def cancel(self):
        # overiding cancel to cleanup self.fig
        # because the call to the destructor __del__ seems to be delayed
        self.plotter.close_fig()
        self.adjuster.close_fig()
        print("ElutionMapperCanvas: closed figs")
        Dialog.cancel(self)

    def show( self ):
        self.toolbar    = True
        fully_automatic = get_setting( 'fully_automatic' )
        if self.mapper.use_mtd_conc:
            if fully_automatic == 1:
                self.parent.after(2000, lambda: self.ok_button.invoke() )
            else:
                pass
        else:
            if fully_automatic == 1:
                self.start_full_automatic_control()
            else:
                auto_navigated_dailog = get_setting( 'auto_navigated_dailog' )
                if self.initial_navigation and auto_navigated_dailog == 2:
                    self.parent.after(100, self.auto_navigate_to_decomp_editor )
        self._show()

    def body(self, body_frame):   # overrides parent class method
        from .ElutionMapperPlotter import ElutionMapperPlotter

        tk_set_icon_portable( self, module=self.caller_module )

        self.upper_frame = Tk.Frame(body_frame)
        self.upper_frame.pack()

        self.plotter    = ElutionMapperPlotter( self, self.upper_frame, self.serial_data, mapper=self.mapper )
        self.plotter.in_range_adjustment = False    # should be changed depending on self.initial_std_diff
        self.plotter.draw()
        self.state_depedents.append( self.plotter.show_diff_btn )

        self.guide_frame = Tk.Frame(body_frame)
        self.guide_frame.pack(fill=Tk.X)
        guide_label = Tk.Label(self.guide_frame, bg='white')
        guide_label.pack(fill=Tk.X, pady=10)
        self.guide_message = guide_label

        self.set_basic_buttons(self.upper_frame)

        self.parent.config(cursor='')

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack(fill=Tk.X, expand=1, padx=20)
        bottom_space = Tk.Frame(self, height=10)
        bottom_space.pack()

        num_buttons = 4
        if is_developing_version():
            num_buttons += 1

        for j in range(num_buttons):
            box.columnconfigure(j, weight=1)

        w = Tk.Button(box, text="◀ Main", command=self.cancel)
        w.grid(row=0, column=0, sticky=Tk.W, padx=10, pady=5)
        self.cancel_button = w

        w = Tk.Button(box, text="▽ Range Editor", command=self.show_range_editor_dialog)
        w.grid(row=0, column=1)
        self.reditor_btn = w
        self.state_depedents.append(w)
        self.default_btn_colors = [ self.reditor_btn.cget(atr) for atr in ['fg', 'bg']]

        state = Tk.DISABLED if self.mapper.get_conc_type() == 2 else Tk.NORMAL

        self.decomp_btn_blink = BlinkingFrame(box)
        self.decomp_btn_blink.grid(row=0, column=2)

        self.decomp_btn = Tk.Button(self.decomp_btn_blink, text="▽ Decomposition Editor", command=self.show_decomp_editor, state=state)
        self.decomp_btn.pack()

        self.decomp_btn_blink.objects = [self.decomp_btn]

        self.state_depedents.append(self.decomp_btn)

        w = Tk.Button(box, text="▶ Execute", command=self.ok, default=Tk.ACTIVE)
        w.grid(row=0, column=3, sticky=Tk.E, padx=10, pady=5)
        self.ok_button = w
        self.state_depedents.append(w)

        if is_developing_version():
            w = Tk.Button(box, text="DevelTest", command=self.do_devel_test)
            w.grid(row=0, column=4, sticky=Tk.E, padx=10, pady=5)

        mapping_canvas_debug = get_setting('mapping_canvas_debug')
        if mapping_canvas_debug:
            w = Tk.Button(box, text="OK Debug", command=self.ok_debug, default=Tk.ACTIVE)
            w.grid(row=0, column=4, sticky=Tk.E, padx=10, pady=5)
            self.ok_debug = w
            self.state_depedents.append(w)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def update_button_colors(self):
        use_elution_models = get_setting('use_elution_models')
        ready_to_execute = False
        if use_elution_models:
            self.reditor_btn.config(fg=self.default_btn_colors[0], bg=self.default_btn_colors[1])
            self.decomp_btn.config(fg='white', bg='green')
            ready_to_execute = True
        else:
            if self.guide_info.decomp_proc_needed():
                self.reditor_btn.config(fg='black', bg='orange')
                self.decomp_btn.config(fg=self.default_btn_colors[0], bg=self.default_btn_colors[1])
                self.decomp_btn_blink.start()
            else:
                self.reditor_btn.config(fg='white', bg='green')
                self.decomp_btn.config(fg=self.default_btn_colors[0], bg=self.default_btn_colors[1])
                ready_to_execute = True

        if ready_to_execute:
            self.ok_button.config(fg='white', bg='green')
        else:
            self.ok_button.config(fg=self.default_btn_colors[0], bg=self.default_btn_colors[1])

    def validate( self ):   # called on "OK" button press
        if self.adjuster.optimize_btn_blink.switch:
            ok = MessageBox.showinfo( "Confirmation",
                    "You have to 'Cancel' to finish because you have not 'Optimize'd the changed parameters.",
                    parent=self )
            return False

        return True

    def ok_debug(self):
        self.ok_debugging = True
        self.ok()

    def apply( self ):  # overrides parent class method
        self.applied = True
        self.final_std_diff = self.current_std_diff
        self.update_params()
        mapping_params = get_mapper_opt_params()
        self.logger.info("memorizing %s as used_mapping_params.", str(mapping_params))
        set_setting('used_mapping_params', mapping_params)
        self.try_image_save()

    def get_cd_colors(self):
        self.judge_holder.get_cd_colors()
        self.update_guide_info()
        return self.judge_holder.get_cd_colors(self.guide_info)

    def update_params( self ):
        analysis_range_info = get_setting('analysis_range_info')
        if analysis_range_info is None:
            set_default_analysis_range_info(self.mapper.x_curve)

        # set_setting also for adjuster items
        adjuster    = self.adjuster
        if adjuster.use_xray_conc or adjuster.use_mtd_conc:
            uv_baseline_opt = 0
            uv_baseline_type = 0
            uv_baseline_adjust = 0
            uv_baseline_with_bpa = None
        else:
            uv_baseline_opt = adjuster.uv_baseline_opt.get()
            uv_baseline_type = adjuster.uv_baseline_type.get()
            uv_baseline_adjust = adjuster.uv_baseline_adjust.get()
            uv_baseline_with_bpa = adjuster.uv_baseline_with_bpa.get()

        set_setting( 'uv_baseline_opt',         uv_baseline_opt )
        set_setting( 'uv_baseline_type',        uv_baseline_type )
        set_setting( 'uv_baseline_adjust',      uv_baseline_adjust )
        set_setting( 'uv_baseline_with_bpa',    uv_baseline_with_bpa )
        set_setting( 'xray_baseline_opt',       adjuster.xray_baseline_opt.get() )
        set_setting( 'xray_baseline_type', adjuster.xray_baseline_type.get() )
        set_setting( 'xray_baseline_adjust',    adjuster.xray_baseline_adjust.get() )
        set_setting( 'xray_baseline_with_bpa',     adjuster.xray_baseline_with_bpa.get() )

        # for backward compatibiliy
        baseline_degree = self.adjuster.get_xray_base_degree()
        set_setting( 'baseline_degree', baseline_degree )
        # scattering_correction will be set in AnalyzerDialog

    def update_params_for_data_correction(self):
        self.update_params()

    def try_image_save( self ):
        test_pattern = get_setting('test_pattern')
        if test_pattern is None or test_pattern < 7:
            return

        mapping_image_folder = get_setting('mapping_image_folder')
        if mapping_image_folder is not None and os.path.exists(mapping_image_folder):
            self.plotter.save_the_figure( mapping_image_folder, get_setting('analysis_name') )

    def get_xray_correction_necessity( self ):
        adjuster    = self.adjuster
        return adjuster.xray_baseline_opt.get() > 0 or adjuster.xray_baseline_adjust.get() == 1

    def update_guide_info(self):
        guide_info = get_analysis_guide_info(self.mapper, self.judge_holder, self.discomfort)
        if guide_info.extra_infos is None:
            # i.e, when "already decomposed"
            # we could replace guide_info now since "already decomposed" info has been separated
            pass
        else:
            self.guide_info = guide_info
        return self.guide_info      # this return value will be used in QuickAnalysis.JudgeHolder

    def get_guide_info( self ):
        self.discomfort = self.serial_data.compute_scattering_baseline_discomfort()
        assert self.initial_std_diff is not None
        self.update_guide_info()
        self.three_d_guide  = self.guide_info.three_d_guide

    def update_guide_message(self):
        self.update_guide_info()
        self.guide_message.config( text=self.guide_info.message, fg=self.guide_info.fg )
        self.logger.info('guide: ' + self.guide_info.message.replace('▶ ', '>>'))  # replace('▶ ', '>>') : to avoid command prompt encode error

    def set_basic_buttons( self, body_frame ):
        self.edit_mode = self.guide_info.edit_mode

        button_bar = Tk.Frame( body_frame )
        button_bar.pack( fill=Tk.X, padx=10, pady=5 )

        self.toggle_button = Tk.Button( button_bar, text=self.get_toggle_text(), command=self.toggle_mode )
        self.toggle_button.pack( side=Tk.LEFT )
        self.state_depedents.append( self.toggle_button )

        self.adjuster   = ElutionMapperAdjuster(body_frame, self)
        self.adjuster.pack( side=Tk.LEFT, fill=Tk.BOTH, expand=1, padx=20 )
        self.state_depedents.append(self.adjuster.manual_sync_btn)

        other_btn_frame = self.plotter.other_btn_frame
        btn_pady = 5
        text = show_button_text_dict[ self.get_toggled_mode() ]
        self.show_button = Tk.Button(other_btn_frame, text=text, width=25,
                                command=self.show_curve, state=Tk.NORMAL )
        self.show_button.pack(side=Tk.RIGHT, padx=10, pady=btn_pady )
        self.state_depedents.append( self.show_button )

        space = Tk.Frame(other_btn_frame, width=self.get_canvas_width()*0.14)
        space.pack(side=Tk.RIGHT)

        investigate_btn = Tk.Button(other_btn_frame, text="Investigate Correction", command=self.investigate_correction)
        investigate_btn.pack(side=Tk.RIGHT, padx=10)
        self.state_depedents.append(investigate_btn)

        pre_sync_btn = Tk.Button(other_btn_frame, text="3D View", command=self.show_threedim_dialog)
        pre_sync_btn.pack(side=Tk.RIGHT, padx=10)
        self.state_depedents.append(pre_sync_btn)

        # work-around to get the layout right; still required?
        self.edit_mode ^= True
        self.after( 100, lambda: self.toggle_mode() )

    def get_edit_mode( self ):
        return self.edit_mode

    def toggle_mode( self ):
        if self.edit_mode:
            self.edit_mode  = False
            self.adjuster.pack_forget()
            self.toggle_button.config( text=self.get_toggle_text() )
        else:
            self.edit_mode  = True
            self.adjuster.pack( fill=Tk.BOTH, expand=Tk.Y, pady=0 )
            self.toggle_button.config( text=self.get_toggle_text() )

    def get_toggle_text( self ):
        text = 'Hide Options' if self.edit_mode else 'Show Options'
        return text

    def change_depedent_states( self, state ):
        for w in self.state_depedents:
            w.config(state=state, fg=self.default_btn_colors[0], bg=self.default_btn_colors[1])

    def draw( self, clear=False, restrict_info=None ):
        self.plotter.draw( clear=clear, restrict_info=restrict_info )

    def draw_mapped( self, clear=False ):
        self.plotter.draw( clear=clear )

    def get_helper_info( self ):
        # temporary fix. to be removed
        # return self.plotter.helper_info
        return None

    def update_helper_info( self, helper_info ):
        # temporary fix. to be removed
        # self.plotter.helper_info        = helper_info
        # self.serial_data.helper_info    = helper_info
        pass

    def get_cursor_widgets( self ):
        return [ self, self.plotter.mpl_canvas_widget ]

    def update_range_info( self, update_list=False ):
        if get_setting( 'range_type' ) == 2:
            manual_range_info = get_setting( 'manual_range_info' )
            if manual_range_info is not None:
                self.parent.range_info = [ manual_range_info ]
            ranges_ = self.parent.range_info[0]
        else:
            ranges_ = self.mapper.get_int_ranges()

        self.range_info     = [ ranges_ ]
        self.annotation_points  = ranges_

        if update_list:
            self.range_changed  = True
            init_ranges = get_init_ranges( just_ranges=ranges_ )
            self.range_list_entry.refresh_spin_boxes( init_ranges )
        else:
            self.range_changed  = False

    def get_range_info( self ):
        analysis_range_info = get_setting('analysis_range_info')
        if analysis_range_info is None:
            return self.range_info
        else:
            return [ analysis_range_info.get_ranges() ]

    def set_adjust_frame( self, bframe ):
        # to avoid for the right column to shrink too small (i.e., smaller than the right pane)
        # self.update()
        self.update_idletasks()
        # self.window_width   = self.winfo_width()
        # get the figure size in pixels in stead because the above does not work properly
        figsize = self.plotter.get_figsize()
        self.window_width   = int( figsize[0] )
        self.right_col_adjust_frame = Tk.Frame( bframe, width=int( self.window_width/3 ), height=0 )
        self.right_col_adjust_frame.grid( row=0, column=2 )

    def get_canvas_width( self ):
        return self.plotter.get_canvas_width()

    def show_absorbance_figure( self ):
        pass

    def show_scattering_figure( self ):
        from .ElutionMapperPlotter import SUGGEST_DISCOMFORT_BY_TEXT
        if self.adjuster.optimize_btn_blink.switch:
            ok = MessageBox.showinfo( "Confirmation",
                    "You have to complete 'Optimize' before investigation\nbecause you have changed optimizer options.",
                    parent=self )
            return False

        # TODO: unify with show_absorbance_figure_util in AbsorbancePlot.py
        print( 'show_scattering_figure' )
        if not SUGGEST_DISCOMFORT_BY_TEXT:
            self.blink_button_frame.stop()

        # TODO: this setting shoud be temporary
        # set_setting( 'correction_iteration', self.adjuster.correction_iteration.get() )

        self.config( cursor='wait' )
        self.update()

        affine_info = self.mapper.get_affine_info()

        if USE_NEW_3D_PLOT_CLASS:
            from molass_legacy.SerialAnalyzer.Scattering3dPlot   import Scattering3dPlotDialog
            data_list = [ self.serial_data.intensity_array ]
            if ENABLE_SIMULATION or get_setting('enable_drift_simulation') == 1:
                if self.three_d_guide:
                    from molass_legacy.SerialAnalyzer.DriftSimulation import DriftSimulation
                    from molass_legacy.SerialAnalyzer.DriftAnalyzer import create_mapper_for_drift_analysis

                    sd  = self.serial_data
                    mapper = create_mapper_for_drift_analysis( self, sd )
                    sim = DriftSimulation( sd, mapper )
                    self.sim_info = [ mapper, sim ]
                    sim_data, sim_base = sim.create_sim_data()
                    data_list.append( sim_data )
                else:
                    sim_base = None
            else:
                sim_base = None
            self.scatterning_plot = dialog = Scattering3dPlotDialog( self, self.serial_data, self.mapper, data_list=data_list, sim_base=sim_base )
            dialog.show()
            if dialog.applied:
                if dialog.apply_fitted_base.get() == 1:
                    set_setting( 'base_drift_params', sim.drift_params )
        else:
            from molass_legacy.SerialAnalyzer.ScatteringPlot import ScatteringPlot
            # for the construction of ScatteringPlot, curve_y should be self.mapper.x_vector
            # i.e., not corrected xray-curve
            self.scatterning_plot = ScatteringPlot( self.qvector, self.mapper.x_vector, self.intensity_array, self.xray_slice, affine_info )
            if self.adjuster is None:
                baseline_degree = None
                do_correction   = False
            else:
                baseline_degree = self.adjuster.get_xray_base_degree()
                # do_correction   = self.adjuster.xray_baseline_opt.get() > 0
                do_correction   = True
            self.scatterning_plot.draw_3d_for_gui( self, 'Xray Scattering from ' + get_setting( 'in_folder' ), do_correction, baseline_degree )
        self.config( cursor='' )

    def get_toggled_mode( self ):
        if self.plotter.mapping_show_mode == 'locally':
            mode = 'uniformly'
        else:
            mode = 'locally'
        return mode

    def show_absorbance_viewer( self ):
        from molass_legacy.UV.AbsorbanceViewer import AbsorbanceViewer
        self.viewer = AbsorbanceViewer( self.serial_data.absorbance, helper_info=self.serial_data.helper_info )
        self.viewer.show( self )

    def show_curve( self ):
        previous_mode = self.plotter.mapping_show_mode
        mode = self.get_toggled_mode()
        self.plotter.mapping_show_mode = mode
        self.plotter.update_show_diff_btn_state()
        text = show_button_text_dict[ previous_mode ]
        self.show_button.config( text=text )
        self.draw_mapped( clear=True )

    def investigate_correction( self ):
        if self.adjuster.optimize_btn_blink.switch:
            ok = MessageBox.showinfo( "Confirmation",
                    "You have to complete 'Optimize' before investigation\nbecause you have changed optimizer options.",
                    parent=self )
            return False

        data = self.intensity_array
        jvector = np.arange( data.shape[0] )
        qvector = data[0,:,0]
        corrected_data = copy.deepcopy( data )
        affine_info = self.mapper.get_affine_info()
        need_adjustment = self.adjuster.xray_baseline_adjust.get()
        baseline_opt = self.adjuster.xray_baseline_opt.get()
        baseline_type = self.adjuster.xray_baseline_type.get()

        from molass_legacy.SerialAnalyzer.ScatteringBaseCorrector import ScatteringBaseCorrector

        corrector = ScatteringBaseCorrector( jvector, qvector, corrected_data,
                                        curve=self.mapper.sd_xray_curve,
                                        affine_info=affine_info,
                                        inty_curve_y=self.mapper.x_vector,
                                        baseline_opt=baseline_opt,
                                        baseline_type=baseline_type,
                                        need_adjustment=need_adjustment,
                                        parent=self, with_demo=True )

        from molass_legacy.SerialAnalyzer.ScatteringBaseInvestigator import InvestigatorDialog
        investigator = InvestigatorDialog( self.parent, corrector )
        investigator.show()

    def show_range_editor_dialog(self, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.RangeEditors.RangeEditorDialog
            reload(molass_legacy.RangeEditors.RangeEditorDialog)
        from molass_legacy.RangeEditors.RangeEditorDialog import RangeEditorDialog

        self.grab_set()     # temporary fix to the grab_release problem
        self.change_depedent_states(Tk.DISABLED)
        self.update_params_for_data_correction()
        self.range_editor = RangeEditorDialog(self.parent, "Range Editor", self.serial_data, self.mapper, self.judge_holder)
        self.range_editor.show()
        self.change_depedent_states(Tk.NORMAL)
        self.update_button_colors()
        self.grab_set()     # temporary fix to the grab_release problem

    def debug_sd(self):
        from molass_legacy.DataStructure.MdViewer import MdViewer
        sd = self.serial_data
        M = sd.intensity_array[:,:,1].T
        mdv = MdViewer(self, M, xvector=sd.qvector)
        mdv.show()

    def show_decomp_editor( self, counter=None, debug=True):
        # import Decomposer
        if debug:
            from importlib import reload
            import molass_legacy.RangeEditors.DecompEditorDialog
            reload(molass_legacy.RangeEditors.DecompEditorDialog)
        from molass_legacy.RangeEditors.DecompEditorDialog import DecompEditorDialog

        self.decomp_btn_blink.stop()
        self.grab_set()     # temporary fix to the grab_release problem
        self.change_depedent_states(Tk.DISABLED)
        self.update_params_for_data_correction()
        # self.serial_data.log_id_values("before DecompEditorDialog")
        self.decomp_editor = DecompEditorDialog(self.parent, "Decomposition Editor", self.serial_data, self.mapper, self.judge_holder)
        if counter is not None:
            counter[0] += 1
            self.update()
        self.decomp_editor.show(counter=counter)
        # self.serial_data.log_id_values("after DecompEditorDialog")
        self.update_guide_message()
        self.change_depedent_states(Tk.NORMAL)
        self.update_button_colors()
        self.grab_set()     # temporary fix to the grab_release problem

    def has_decomp_editor(self):
        return self.decomp_editor is not None

    def auto_navigate_to_decomp_editor( self ):
        counter = [0]
        def progress_cb():
            self.update()
            return counter[0]

        def show_progress():
            def navigate():
                counter[0] += 1
                self.update()
                """
                adjuster = self.adjuster
                if not adjuster.get_edit_mode():
                    adjuster.toggle_button.invoke()
                    self.update()
                if adjuster.get_conc_type() == 0:
                    adjuster.uv_baseline_adjust.set( 1 )
                adjuster.xray_radio_buttons[1].invoke()
                adjuster.xray_baseline_adjust.set( 1 )
                counter[0] += 1
                self.update()
                adjuster.optimize_btn.invoke()
                counter[0] += 1
                self.update()
                """
                self.show_decomp_editor(counter=counter)

            self.after(100, navigate)
            from molass_legacy.KekLib.ProgressMinDialog import ProgressMinDialog
            counter[0] += 1
            for k in range(3):
                try:
                    self.progress = ProgressMinDialog(self,
                                title="Navigation Progress",
                                message="Please be patient until the Decomposition Editor appears.",
                                length=240,
                                num_steps=6, progress_cb=progress_cb, interval=100)
                    break
                except:
                    # this case may occur when the machine is buzy for running anti-virus software.
                    # better solution is desired.
                    self.logger.warning("failed to show Navigation Progress. retry after 1 sec.")
                    self.update()
                    time.sleep(1)

        self.after(100, show_progress)

    def start_full_automatic_control( self ):
        if self.guide_info.decomp_proc_needed():     # to be consistent with the Tester
            self.parent.after(100, self.auto_navigate_to_decomp_editor)
            self.parent.after(1000, self.start_full_automatic_control_impl)
        else:
            self.parent.after(100, self.auto_navigate_no_decomp)

    def start_full_automatic_control_impl( self ):
        self.control_thread = threading.Thread(
                                target=self.control_gui_automatically,
                                name='GuiControlThread',
                                args=[]
                                )

        self.control_queue = queue.Queue()
        self.control_thread.start()
        self.after(1000, self.monitor_control_thread)

    def monitor_control_thread( self ):
        try:
            ret = self.control_queue.get(block=False)
            self.control_thread.join()
            time.sleep(1)
            self.ok_button.invoke()
        except Exception as exc:
            print("monitor_control_thread", exc)
            self.parent.after(1000, self.monitor_control_thread)

    def decomp_editor_state( self ):
        try:
            return self.decomp_editor.is_ready()
        except:
            print( 'no self.decomp_editor' )
        return False

    def control_gui_automatically( self ):
        print("control_gui_automatically")
        while not self.decomp_editor_state():
            print( "wainting for decomp_editor being ready" )
            time.sleep(1)

        time.sleep(1)
        self.decomp_editor.ok_button.invoke()
        self.control_queue.put(1)

    def auto_navigate_no_decomp(self):
        self.ok_button.invoke()

    def show_threedim_dialog(self):
        self.dialog.show_threedim_dialog()

    def get_sd(self):
        return self.serial_data

    def do_devel_test(self):
        from importlib import reload
        import molass_legacy.Tools.EmbedCushion
        reload(molass_legacy.Tools.EmbedCushion)
        from molass_legacy.Tools.EmbedCushion import embed_cushion

        embed_cushion(self)



"""
    Kratky.GuinierKratkyView.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import os
import numpy as np
from scipy.stats import linregress
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from GuinierTools.RgCurveUtils import plot_rg_curves
from molass_legacy._MOLASS.SerialSettings import get_setting

PLOT_EOII_IN_GUINIER = False
TOGGLE_BUTTON_TEXTS = ["Show", "Hide"]

class GuinierKratkyView(Dialog):
    def __init__(self, parent, optimizer, params, lrf_info):
        self.parent = parent
        self.optimizer = optimizer
        self.params = params
        self.lrf_info = lrf_info
        self.separate_eoii = get_setting("separate_eoii")
        self.show_baseline = 0
        self.all_axes = []

        Dialog.__init__(self, parent, "Guinier-Kratky View", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()

        # fig, axes = plt.subplots(ncols=4, figsize=(20,5), subplot_kw=dict(projection="3d"))
        fig = plt.figure(figsize=(18,9))

        self.fig = fig
        self.draw()
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.draw()

    def buttonbox(self):
        lower_frame = Tk.Frame(self)
        lower_frame.pack(fill=Tk.BOTH, expand=1)

        width = int(self.mpl_canvas_widget.cget('width'))
        padx = width*0.05

        tframe = Tk.Frame(lower_frame)
        tframe.pack(side=Tk.LEFT, padx=padx)
        self.toolbar = NavigationToolbar( self.mpl_canvas, tframe )
        self.toolbar.update()

        space = Tk.Frame(lower_frame, width=padx)
        space.pack(side=Tk.RIGHT)

        box = Tk.Frame(lower_frame)
        box.pack(side=Tk.RIGHT)

        if False:
            w = Tk.Button(box, text="Tutorial Figure", width=15, command=self.show_tutorial_figure)
            w.pack(side=Tk.LEFT, pady=5)

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=padx, pady=5)

        text = self.get_toggle_button_text()
        w = Tk.Button(box, text=text, width=12, command=self.toggle_baseline)
        w.pack(side=Tk.LEFT, padx=padx, pady=5)
        self.toggle_btn = w

        w = Tk.Button(box, text="Show Reducibility", width=15, command=self.show_reducibility)
        w.pack(side=Tk.LEFT, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def get_toggle_button_text(self):
        return "%s Baseline" % TOGGLE_BUTTON_TEXTS[self.show_baseline]

    def draw(self, debug=True):
        from molass_legacy.DataStructure.MatrixData import simple_plot_3d
        from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
        if debug:
            from importlib import reload
            import Kratky.GuinierKratkyInfo
            reload(Kratky.GuinierKratkyInfo)
        from Kratky.GuinierKratkyInfo import GuinierKratkyInfo

        lrf_info = self.lrf_info
        gk_info = GuinierKratkyInfo(self.optimizer, self.params, lrf_info, self.show_baseline)
        self.gk_info = gk_info

        rg_params = gk_info.rg_params       # got valid_rgs
        points = lrf_info.estimate_xr_peak_points()
        heights = points[:,1]
        ip, iq = sorted(np.argpartition(heights,-2)[-2:])
        ib = len(rg_params)

        Pxr, Cxr, Puv, Cuv  = lrf_info.matrices[:4]

        fig = self.fig
        gs = GridSpec(2,12)

        self.clear_axes()

        axes = []
        for i in range(3):
            ax = fig.add_subplot(gs[0,i*4:(i+1)*4])
            axes.append(ax)
            self.all_axes.append(ax)
        ax1_, ax2_, ax3_ = axes

        ax1_.set_title("Xray Elution Decomposition", fontsize=16)
        ax1_.set_xlabel("Eno")
        ax1_.set_ylabel("Intensity")

        nc = len(rg_params)
        x = lrf_info.x
        y = lrf_info.y
        ax1_.plot(x, y, color="orange")
        xr_ty = np.zeros(len(x))
        xr_cy_list = lrf_info.get_xr_cy_list()
        for k, cy in enumerate(xr_cy_list):
            if k == nc:
                color = "red"
                label = "baseline"
            else:
                color = None
                label = "component-%d" % (k+1)
            if k < nc or self.show_baseline:
                ax1_.plot(x, cy, ":", color=color, label=label)

        if self.separate_eoii:
            cy = Cxr[-1,:]
            ax1_.plot(x, cy, ":", color="pink", label="effect of ii")

        ax1_.legend()

        axt = ax1_.twinx()
        self.all_axes.append(axt)
        axt.grid(False)

        xr_ty = lrf_info.xr_ty
        rg_curve = self.optimizer.rg_curve
        plot_rg_curves(axt, heights, rg_params, x, xr_cy_list, xr_ty, rg_curve)

        ax2_.set_title("Xray Component Scattering (Guinier Plot)", fontsize=16)
        ax2_.set_xlabel("$Q^2$")
        ax2_.set_ylabel(r"$L_n(I) - L_n(I_0)$")

        ax3_.set_title("Xray Component Scattering (Kratky Plot)", fontsize=16)
        ax3_.set_xlabel("$QR_g$")
        ax3_.set_ylabel(r"$(QR_g)^2 \times I(Q)/I(0)$")

        rg_ = np.average(rg_params)

        gslice = gk_info.gslice
        qv2 = gk_info.qv2
        qv2_ = gk_info.qv2_
        glny_s = gk_info.glny_s
        qrgs = gk_info.qrgs
        qrgnys = gk_info.qrgnys
        # masses = gk_info.get_molecular_masses()

        for k, cy in enumerate(Pxr.T[0:nc+1]):
            if k == nc:
                color = "red"
                label_g = "baseline"
                label_k = label_g
            else:
                color = None
                label_g = "component-%d, $R_g=%.1f$" % (k+1, gk_info.rgs[k])    # note that gk_info.rgs[k] != gk_info.rg_params[k]
                label_k = "component-%d" % (k+1)

            if k < nc or self.show_baseline:
                glny_ = glny_s[k]
                ax2_.plot(qv2_, glny_, ":", color=color, label=label_g)
                qrg = qrgs[k]
                qrgny = qrgnys[k]
                ax3_.plot(qrg, qrgny, 'o', markersize=1, color=color, label=label_k)

        xmin, xmax = ax3_.get_xlim()
        ymin, ymax = ax3_.get_ylim()
        ax3_.set_xlim(xmin, xmax)
        ax3_.set_ylim(ymin, ymax)
        px = np.sqrt(3)
        py = 3/np.e
        ax3_.plot([px, px], [ymin, ymax], ":", color="gray")
        ax3_.plot([xmin, xmax], [py, py], ":", color="gray")

        dy = (ymax - ymin)*0.01
        ax3_.text(px, ymin+dy, r"$ \sqrt{3} $", ha="right")
        ax3_.text(xmax, py+2*dy, r"$ 3/e $", ha="right")

        ax3_.plot([xmin, xmax], [0, 0], color="red", alpha=0.5)

        if PLOT_EOII_IN_GUINIER:
            if self.separate_eoii:
                cy = Pxr[:,-1]
                ax2_.plot(qv2_, np.log(cy[gslice]), ":", color="pink", label="effect of ii")

        ax2_.legend()
        # ax3_.legend(bbox_to_anchor=(0.5, 1), loc="upper center", borderaxespad=2)
        ax3_.legend(loc="upper center")

        axes = []
        for i in range(4):
            ax = fig.add_subplot(gs[1,i*3:(i+1)*3], projection="3d")
            axes.append(ax)
            self.all_axes.append(ax)

        ax0, ax1, ax2, ax3 = axes

        in_folder = get_in_folder()
        fig.suptitle("LRF into Components and a Baseline from %s" % in_folder, fontsize=24)

        ax0.set_title(r"$M = P \cdot C + Residual$")

        D = self.optimizer.xrD
        qv = self.optimizer.qvector
        simple_plot_3d(ax0, D, x=qv)

        zlim = ax0.get_zlim()

        ax1.set_title(r"$p_{0} \cdot c_{0}$".format(ip+1), fontsize=16)
        D1  = Pxr[:,[ip]] @ Cxr[[ip],:]
        simple_plot_3d(ax1, D1, x=qv)
        ax1.set_zlim(zlim)

        ax2.set_title(r"$p_{0} \cdot c_{0}$".format(iq+1), fontsize=16)
        D2  = Pxr[:,[iq]] @ Cxr[[iq],:]
        simple_plot_3d(ax2, D2, x=qv)
        ax2.set_zlim(zlim)

        ax3.set_title(r"$p_b \cdot c_b$", fontsize=16)
        D3  = Pxr[:,[ib]] @ Cxr[[ib],:]
        simple_plot_3d(ax3, D3, x=qv)
        ax3.set_zlim(zlim)

        fig.tight_layout()
        # fig.subplots_adjust(top=0.9, hspace=0.8, bottom=0.04)

    def clear_axes(self):
        for ax in self.all_axes:
            ax.remove()
        self.all_axes = []

    def toggle_baseline(self):
        self.show_baseline = 1 - self.show_baseline
        self.draw()
        self.mpl_canvas.draw()
        text = self.get_toggle_button_text()
        self.toggle_btn.config(text=text)
        self.update()

    def show_tutorial_figure(self, debug=True):
        if debug:
            from importlib import reload
            import Optimizer.SvdTutorial
            reload(Optimizer.SvdTutorial)
        from molass_legacy.Optimizer.SvdTutorial import SvdTutorial

        dialog = SvdTutorial(self.parent, self)
        dialog.show()

    def show_reducibility(self):
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        from molass_legacy.Optimizer.ElutionComposer import make_composites_from_deviations
        from molass_legacy.Optimizer.CompositeInfo import convert_for_display

        ratios = self.gk_info.compute_adjacent_deviation_ratios()
        composites = make_composites_from_deviations(ratios)

        precision_save = np.get_printoptions()["precision"]
        np.set_printoptions(precision=2)
        ratios_str = str(ratios)
        np.set_printoptions(precision=precision_save)
        composites_str = str(convert_for_display(composites))

        MessageBox.showinfo( "Reducibility Info",
            "Adjacent Deviation Ratios: %s\n"
            "Possible Reduction: %s"
            % (ratios_str, composites_str),
            parent=self)
